import atexit
import hashlib
import logging
import os
import tempfile
import urllib.parse
import urllib.request
import weakref

import numpy
from skimage.exposure import rescale_intensity
from skimage.util import img_as_float32

import cellprofiler_core.preferences
from .._abstract_image import AbstractImage
from ..._image import Image
from ....reader import get_image_reader, get_image_reader_class
from ....utilities.image import is_numpy_file, download_to_temp_file
from ....utilities.image import is_matlab_file
from ....utilities.image import loadmat
from ....utilities.image import load_data_file
from ....utilities.image import generate_presigned_url
from ....constants.image import FILE_SCHEME, PASSTHROUGH_SCHEMES, NO_RESCALE
from ....utilities.pathname import pathname2url, url2pathname

LOGGER = logging.getLogger(__name__)

# A set of readers with open file locks.
ACTIVE_READERS = weakref.WeakSet()


class FileImage(AbstractImage):
    """Base for image providers: handle pathname and filename & URLs"""

    def __init__(
        self,
        name,
        pathname,
        filename,
        rescale_range=None, # autoscale by default
        metadata_rescale=False,
        series=None,
        index=None,
        channel=None,
        volume=False,
        spacing=None,
        z=None,
        t=None
    ):
        """
        :param name: Name of image to be provided
        :type name:
        :param pathname: Path to file or base of URL
        :type pathname:
        :param filename: Filename of file or last chunk of URL
        :type filename:
        :param rescale_range: a 2-tuple of min/max float values dictating the values to manually rescale the image from
                              if `None`, autorescaling will occur by dtype (if `metadata_rescale` is `False`)
                              or by values dictated in the image fle header's metadata (if `metadata_rescale` is `True`)
                              if (None, None) is given, no rescaling will occur (see NO_RESCALE constant), regardless of
                              image values
        :type rescale_range: (float, float)
        :param metadata_rescale: If `True`, rescale the image by the bitdepth values located in image file header's metadata
                                 (useful for e.g. 12-bit images stored in float16 dtype)
        :type metadata_rescale: bool
        :param series:
        :type series:
        :param index:
        :type index:
        :param channel:
        :type channel:
        :param volume:
        :type volume:
        :param spacing:
        :type spacing:
        """
        if pathname.startswith(FILE_SCHEME):
            pathname = url2pathname(pathname)
        self.__name = name
        self.__pathname = pathname
        self.__filename = filename
        self.__cached_file = None
        self.__is_cached = False
        self.__cacheing_tried = False
        self.__image = None
        self.__preferred_reader = None
        self.__image_file = None
        self.__reader = None

        if pathname is None:
            self.__url = filename
        elif any([pathname.startswith(s + ":") for s in PASSTHROUGH_SCHEMES]):
            if filename is not None:
                self.__url = pathname + "/" + filename
            else:
                self.__url = pathname
        elif filename is None:
            self.__url = pathname2url(pathname)
        else:
            self.__url = pathname2url(os.path.join(pathname, filename))
        self.__scale = None
        if rescale_range != None and metadata_rescale != False:
            LOGGER.error("Cannot specify both rescale_range and metadata_rescale, prioritizing rescale_range")
            metadata_rescale = False
        self.__rescale_range = rescale_range
        self.__metadata_rescale = metadata_rescale
        self.__series = series
        self.__channel = channel
        self.__index = index
        self.__volume = volume
        self.__spacing = spacing
        if volume:
            if z is not None and t is not None:
                raise ValueError(f"T- and Z-plane indexes were specified while in 3D mode."
                                 f"If extracting planes you should disable separation of an axis to "
                                 f"work in 3D.")
            self.__z = z
            self.__t = t
        else:
            self.__z = z if z is not None else 0
            self.__t = t if t is not None else 0

    @staticmethod
    def __validate_rescale_range(rescale_range):
        if not isinstance(rescale_range, tuple):
            raise ValueError(f"rescale_range must be a tuple, got {type(rescale_range)}")
        if len(rescale_range) != 2:
            raise ValueError("rescale_range must be a tuple of length 2")
        if not isinstance(rescale_range[0], (float, int)):
            raise ValueError("rescale_range must be a tuple of floats")
        if not isinstance(rescale_range[1], (float, int)):
            raise ValueError("rescale_range must be a tuple of floats")
        if rescale_range[0] > rescale_range[1]:
            raise ValueError("rescale_range must be min-max")

    @staticmethod
    def __uint_to_float32(data, wants_inscale=False):
        scale = float(2**(8*data.dtype.itemsize) - 1)
        # out_range set to float64 seems wasteful for small types like
        # [u]int[8,16] (and it is), but it gets us a bit of extra precision
        # before a final cast to float32, and the internals of
        # rescale_intensity do it *anyway* (it does np.clip, w/ min/max values
        # set to float64, causing the img to become float64) so we don't
        # incur *extra* inefficiencies on top of what skimage is already doing
        data = rescale_intensity(data, in_range="dtype", out_range="float64").astype("float32")
        if wants_inscale:
            return data, scale
        return data

    @staticmethod
    def __int_to_float32(data, wants_inscale=False):
        scale = float(2**(8*data.dtype.itemsize) - 1)
        data = rescale_intensity(data, in_range="dtype", out_range=(0., 1.)).astype("float32")
        if wants_inscale:
            return data, scale
        return data

    @staticmethod
    def __float_to_float32(data, wants_inscale=False):
        # data is normalized float, just cast to 32 bit
        if data.min() >= 0 and data.max() <= 1:
            scale = 1.0
            data = img_as_float32(data)
        # data is normalized float, but not in [0, 1]
        # adjust and cast to 32 bit
        elif data.min() >= -1 and data.max() <= 1:
            scale = 2.0
            # a bit nasty to cast to 64,
            # but it buys us some extra precision until final cast to 32
            data = (data.astype("float64") + 1) / 2.
            data = img_as_float32(data)
        # data is unnormalized, with non-negative values
        # treat as uint of the same bitdepth and normalize
        elif data.min() >= 0:
            scale = float(2**(8*data.dtype.itemsize)-1)
            data = FileImage.__uint_to_float32(data.astype(f"u{data.dtype.itemsize}"))
        # data is unnormalized, with negative values
        # treat as int of the same bitdepth and normalize
        else:
            data, scale = FileImage.__int_to_float32(
                data.astype(f"i{data.dtype.itemsize}"),
                wants_inscale=True
            )

        if wants_inscale:
            return data, scale
        return data

    @staticmethod
    def normalize_to_float32(data, in_range=None, wants_inscale=False):
        '''
        Convert an array of given data of some type to float32,
        normalizing it to the range 0-1.

        The scaling IS NOT done such that the data's minimal pixel value
        is forced to 0. and the maximal pixel value is forced to 1.
        (i.e. stretching/shrinking the data).
        Instead, scaling IS done such that relative ranges are preserved
        (i.e. range of the data wrt range of the data type is kept).

        For example, for unsigned type, the scaling *will not* be
        data / (max_val - min_val)
        but rather
        data / (2^bit_depth - 1)

        The conversion also handles signed types by shifting the data first.

        Finally the conversion handles three conditions present in floating
        point images:
        1. Floating point images which already have values inside the range 0-1.
           In this case, it is simply cast to float32.
        2. Floating point images which have values normalized inside the range -1-1.
           In this case, the data is renormalized to the range 0-1, and then cast to float32.
        3. Floating point values which *represent* signed/unsigned integers but stored as float
           e.g. ..., -65535.0, -255.0, -1.0, 0.0, 1.0, 65535.0, ...
           i.e. data that really should be an signed/unsigned integer type, but for whatever
           reason, are stored as float type.
           In this case, it is converted to the appropriate integer type first,
           and then normalized and cast to float32 in the range 0-1.

        param: data - the numpy data to be converted to rescaled and cast as float32
        param: in_range - `None` for automatic range detection, or a tuple of (min, max)
                          dictating the range of values in the input data
                          if equal to (None, None) no rescaling or type converstion will be performed
        param: wants_inscale - if True, return the rescaled data, and scale factor used to rescale it
        '''

        if in_range is not None:
            if in_range == NO_RESCALE:
                scale = 1.0
                # do not scale or typecast data
            if in_range != NO_RESCALE:
                scale = float(in_range[1] - in_range[0])
                # see notes above about casting to float64
                data = rescale_intensity(data.astype("float64"), in_range=in_range, out_range=(0., 1.)).astype("float32")
        elif numpy.issubdtype(data.dtype, numpy.floating):
            data, scale = FileImage.__float_to_float32(data, wants_inscale=True)
        elif numpy.issubdtype(data.dtype, numpy.signedinteger):
            data, scale = FileImage.__int_to_float32(data, wants_inscale=True)
        elif numpy.issubdtype(data.dtype, numpy.unsignedinteger):
            data, scale = FileImage.__uint_to_float32(data, wants_inscale=True)
        else:
            raise ValueError(f"Unsupported data type: {data.dtype}")

        if wants_inscale:
            return data, scale
        return data

    @property
    def series(self):
        return self.__series

    @series.setter
    def series(self, series):
        # Invalidate the cached image
        self.__image = None
        self.__series = series

    @property
    def channel(self):
        return self.__channel

    @channel.setter
    def channel(self, index):
        # Invalidate the cached image
        self.__image = None
        self.__channel = index

    @property
    def index(self):
        return self.__index

    @index.setter
    def index(self, index):
        # Invalidate the cached image
        self.__image = None
        self.__index = index

    @property
    def z(self):
        return self.__z

    @property
    def t(self):
        return self.__t

    @property
    def scale(self):
        return self.__scale

    @scale.setter
    def scale(self, scale):
        if self.__scale is not None:
            LOGGER.warning(f"Setting scale to {scale} but scale was {self.__scale}")
        if scale is None:
            LOGGER.warning(f"Setting scale to None, but scale was {self.__scale}")
            self.__scale = scale
        elif isinstance(scale, int):
            self.__scale = float(scale)
        elif isinstance(scale, float):
            self.__scale = scale
        else:
            raise ValueError("scale must be a float or int")

    @property
    def rescale_range(self):
        return self.__rescale_range

    @rescale_range.setter
    def rescale_range(self, rescale_range):
        if rescale_range != None:
            FileImage.__validate_rescale_range(rescale_range)

        if (self.__rescale_range != None):
            LOGGER.warning(f"Rescale range was {str(self.__rescale_range)}, now {str(rescale_range)}")
        self.__rescale_range = rescale_range


    @property
    def metadata_rescale(self):
        return self.__metadata_rescale

    @metadata_rescale.setter
    def metadata_rescale(self, metadata_rescale):
        if not isinstance(metadata_rescale, bool):
            raise ValueError("metadata_rescale must be a boolean")
        self.__metadata_rescale = metadata_rescale

    def get_reader(self, create=True, volume=False):
        if self.__reader is None and create:
            image_file = self.get_image_file()
            self.__reader = get_image_reader(image_file, volume=volume)
            ACTIVE_READERS.add(self)
        return self.__reader

    def get_name(self):
        return self.__name

    def get_pathname(self):
        return self.__pathname

    def get_filename(self):
        return self.__filename

    def cache_file(self):
        """Cache a file that needs to be HTTP downloaded

        Return True if the file has been cached
        """
        if self.__cacheing_tried:
            return self.__is_cached
        self.__cacheing_tried = True
        #
        # Check to see if the pathname can be accessed as a directory
        # If so, handle normally
        #
        path = self.get_pathname()
        if len(path) == 0:
            filename = self.get_filename()
            if os.path.exists(filename):
                return False
            parsed_path = urllib.parse.urlparse(filename)
            url = filename
            if len(parsed_path.scheme) < 2:
                raise IOError("Test for access to file failed. File: %s" % filename)
        elif os.path.exists(path):
            return False
        else:
            parsed_path = urllib.parse.urlparse(path)
            url = "/".join((path, self.get_filename()))
            #
            # Scheme length == 0 means no scheme
            # Scheme length == 1 - probably DOS drive letter
            #
            if len(parsed_path.scheme) < 2:
                raise IOError(
                    "Test for access to directory failed. Directory: %s" % path
                )
        if parsed_path.scheme == "file":
            self.__cached_file = url2pathname(path)
        elif is_numpy_file(self.__filename):
            #
            # urlretrieve uses the suffix of the path component of the URL
            # to name the temporary file, so we replicate that behavior
            #
            temp_dir = cellprofiler_core.preferences.get_temporary_directory()
            tempfd, temppath = tempfile.mkstemp(suffix=".npy", dir=temp_dir)
            self.__cached_file = temppath
            try:
                url = generate_presigned_url(url)
                self.__cached_file, headers = urllib.request.urlretrieve(
                    url, filename=temppath
                )
            finally:
                os.close(tempfd)
        else:
            from ....pipeline import ImageFile
            image_file = self.get_image_file()
            rdr_class = get_image_reader_class(image_file, volume=self.__volume)
            if not rdr_class.supports_url() and parsed_path.scheme.lower() != 'omero':
                cached_file = download_to_temp_file(image_file.url)
                if cached_file is None:
                    return False
                self.__cached_file = pathname2url(cached_file)
                self.__image_file = ImageFile(self.__cached_file)
            else:
                return False
        self.__is_cached = True
        return True

    def get_full_name(self):
        self.cache_file()
        if self.__is_cached:
            return self.__cached_file
        return os.path.join(self.get_pathname(), self.get_filename())

    def get_url(self):
        """Get the URL representation of the file location"""
        return self.__url

    def get_md5_hash(self, measurements):
        """Compute the MD5 hash of the underlying file or use cached value

        measurements - backup for case where MD5 is calculated on image data
                       directly retrieved from URL
        """
        #
        # Cache the MD5 hash on the image reader
        #
        if is_matlab_file(self.__filename) or is_numpy_file(self.__filename):
            rdr = None
        else:
            rdr = self.get_reader()

        if rdr is None or not hasattr(rdr, "md5_hash"):
            hasher = hashlib.md5()
            path = self.get_full_name()
            if not os.path.isfile(path):
                # No file here - hash the image
                image = self.provide_image(measurements)
                hasher.update(image.pixel_data.tostring())
            else:
                with open(self.get_full_name(), "rb") as fd:
                    while True:
                        buf = fd.read(65536)
                        if len(buf) == 0:
                            break
                        hasher.update(buf)
            if rdr is None:
                return hasher.hexdigest()
            # TODO - LIS: rdr.md5_hash is unused
            rdr.md5_hash = hasher.hexdigest()
        return rdr.md5_hash

    def release_memory(self):
        """Release any image memory

        Possibly delete the temporary file"""
        if self.__is_cached:
            if is_matlab_file(self.__filename) or is_numpy_file(self.__filename):
                try:
                    os.remove(self.__cached_file)
                except:
                    LOGGER.warning(
                        "Could not delete file %s", self.__cached_file, exc_info=True
                    )
            self.__is_cached = False
            self.__cacheing_tried = False
            self.__cached_file = None

        rdr = self.get_reader(create=False)
        if rdr is not None:
            rdr.close()
            self.__reader = None
            if self in ACTIVE_READERS:
                ACTIVE_READERS.remove(self)
        self.__image = None

    def __del__(self):
        # using __del__ is all kinds of bad, but we need to remove the
        # files to keep the system from filling up.
        self.release_memory()

    def get_image_file(self):
        if self.__image_file is None:
            from ....pipeline import ImageFile
            self.__image_file = ImageFile(self.get_url())
            self.__image_file.preferred_reader = self.__preferred_reader
        return self.__image_file

    def __set_image(self):
        if self.__volume:
            self.__set_image_volume()
            return

        self.cache_file()
        channel_names = []

        # .mat and .npy are special cases, intended for flatfield illumination correction images
        # we do not want to rescale them
        if is_matlab_file(self.__filename):
            img = load_data_file(self.get_full_name(), loadmat)
            self.rescale_range = NO_RESCALE
            self.__scale = 1.0
        elif is_numpy_file(self.__filename):
            img = load_data_file(self.get_full_name(), numpy.load)
            self.rescale_range = NO_RESCALE
            self.__scale = 1.0
        else:
            rdr = self.get_reader()
            _rescale_range = None
            if numpy.isscalar(self.index) or self.index is None:
                img, _rescale_range = rdr.read(
                    wants_metadata_rescale=True,
                    c=self.channel,
                    z=self.z,
                    t=self.t,
                    series=self.series,
                    index=self.index,
                    channel_names=channel_names,
                )
            else:
                # It's a stack
                stack = []
                if numpy.isscalar(self.series):
                    series_list = [self.series] * len(self.index)
                else:
                    series_list = self.series
                if not numpy.isscalar(self.channel):
                    channel_list = [self.channel] * len(self.index)
                else:
                    channel_list = self.channel
                for series, index, channel in zip(
                    series_list, self.index, channel_list
                ):
                    img, _rescale_range = rdr.read(
                        wants_metadata_rescale=True,
                        c=channel,
                        z=self.z,
                        t=self.t,
                        series=series,
                        index=index,
                        channel_names=channel_names,
                    )
                    stack.append(img)
                img = numpy.dstack(stack)

            # if rescale_range is not manually set in __init__
            # then set it from the metadata, if requested
            if self.rescale_range is None and self.metadata_rescale:
                # _rescale_range might stil be None (reader didn't find anything)
                # but that's okay
                self.rescale_range = _rescale_range

            img, self.__scale = FileImage.normalize_to_float32(img, in_range=self.rescale_range, wants_inscale=True)

        self.__image = Image(
            img,
            path_name=self.get_pathname(),
            file_name=self.get_filename(),
            scale=self.__scale,
            channelstack=img.ndim == 3 and img.shape[-1]>3
        )

        if img.ndim == 3 and len(channel_names) == img.shape[2]:
            self.__image.channel_names = list(channel_names)

    def provide_image(self, image_set):
        """Load an image from a pathname
        """
        if self.__image is None:
            self.__set_image()
        return self.__image

    def __set_image_volume(self):
        pathname = url2pathname(self.get_url())

        # Volume loading is currently limited to tiffs/numpy files only
        if is_numpy_file(self.__filename):
            img = numpy.load(pathname)
        else:
            reader = self.get_reader(volume=True)
            img, self.rescale_range = reader.read_volume(
                wants_metadata_rescale=True,
                c=self.channel,
                z=self.z,
                t=self.t,
                series=self.series
            )

        if self.metadata_rescale:
            img, self.__scale = FileImage.normalize_to_float32(img, in_range=self.rescale_range, wants_inscale=True)
        else:
            img, self.__scale = FileImage.normalize_to_float32(img, wants_inscale=True)

        self.__image = Image(
            image=img,
            path_name=self.get_pathname(),
            file_name=self.get_filename(),
            dimensions=3,
            scale=self.__scale,
            spacing=self.__spacing,
        )


@atexit.register
def shut_down_readers():
    """
    Ensures that reader file locks are released before we shut down CP.
    This isn't a problem on UNIX, but on Windows any cached files can't be
    deleted if they're still linked to a reader instance.
    """
    for reader in list(ACTIVE_READERS):
        reader.release_memory()



from .._event import Event


class RunException(Event):
    """An exception was caught during a pipeline run

    Initializer:
    error - exception that was thrown
    module - module that was executing
    tb - traceback at time of exception, e.g from sys.exc_info
    """

    def __init__(self, error, module, tb=None):
        self.error = error
        self.cancel_run = True
        self.skip_thisset = False
        self.module = module
        self.tb = tb

    def event_type(self):
        return "Pipeline run exception"



import imageio
import numpy

from .... import Image
from .....constants.image import MD_SIZE_C, MD_SIZE_Z, MD_SIZE_T
from .....utilities.image import convert_image_to_objects
from .....utilities.pathname import url2pathname
from ._url_image import URLImage


class ObjectsImage(URLImage):
    """Provide a multi-plane integer image, interpreting an image file as objects"""

    def __init__(self, name, url, series, index, volume=False, spacing=None, z=None, t=None):
        self.__data = None
        self.volume = volume
        if volume:
            index = self.get_indexes(url)
            series = None
        self.__image = None
        self.__spacing = spacing
        URLImage.__init__(
            self,
            name,
            url,
            rescale_range=None,
            metadata_rescale=False,
            series=series,
            index=index,
            volume=volume,
            z=z,
            t=t
        )

    def provide_image(self, image_set):
        """Load an image from a pathname
        """

        if self.__image is not None:
            return self.__image

        if self.volume:
            return self.get_image_volume()

        self.cache_file()
        filename = self.get_filename()
        channel_names = []
        url = self.get_url()
        properties = {}
        if self.index is None:
            reader = self.get_reader()
            meta = reader.get_series_metadata()
            series = self.series
            if self.series is None:
                series = 0
            nplanes = meta[MD_SIZE_C][series] * meta[MD_SIZE_Z][series] * meta[MD_SIZE_T][series]
            indexes = list(range(nplanes))
        elif numpy.isscalar(self.index):
            indexes = [self.index]
        else:
            indexes = self.index
        planes = []
        offset = 0
        for i, index in enumerate(indexes):
            properties["index"] = index
            if self.series is not None:
                if numpy.isscalar(self.series):
                    properties["series"] = self.series
                else:
                    properties["series"] = self.series[i]
            rdr = self.get_reader()
            img = rdr.read(**properties).astype(int)
            img = convert_image_to_objects(img).astype(numpy.int32)
            img[img != 0] += offset
            offset += numpy.max(img)
            planes.append(img)

        image = Image(
            numpy.dstack(planes),
            path_name=self.get_pathname(),
            file_name=self.get_filename(),
            convert=False,
        )
        self.__image = image
        return image

    def get_indexes(self, url):
        pathname = url2pathname(url)
        # Javabridge gave us dud indexes, let's find our own planes
        self.__data = imageio.volread(pathname).astype(int)
        indexes = list(range(self.__data.shape[0]))
        return indexes

    def get_image_volume(self):
        imdata = self.__data
        planes = []
        for planeid in range(imdata.shape[0]):
            planes.append(imdata[planeid].astype(numpy.int32))
        imdata = numpy.stack(planes)

        image = Image(
            imdata,
            path_name=self.get_pathname(),
            file_name=self.get_filename(),
            convert=False,
            dimensions=3,
        )
        self.__image = image
        return image



from .._setting import Setting


class Text(Setting):
    """A setting that displays as an edit box, accepting a string

    """

    def __init__(self, text, value, *args, **kwargs):
        kwargs = kwargs.copy()
        self.multiline_display = kwargs.pop("multiline", False)
        self.metadata_display = kwargs.pop("metadata", False)
        super(Text, self).__init__(text, value, *args, **kwargs)



from ._event import Event


class ModuleAdded(Event):
    """A module was added to the pipeline

    """

    def __init__(self, module_num, is_image_set_modification=False):
        super(ModuleAdded, self).__init__(
            is_pipeline_modification=True,
            is_image_set_modification=is_image_set_modification,
        )
        self.module_num = module_num

    def event_type(self):
        return "Module Added"



import sys

from ..module._plugin_importer import PluginImporter

from ..utilities.core.modules import fill_modules
from ..utilities.core.plugins import load_plugins

sys.meta_path.append(PluginImporter())

fill_modules()
load_plugins()



from cellprofiler_core.constants.measurement import (
    C_URL,
    C_FILE_NAME,
    C_PATH_NAME,
    C_OBJECTS_URL,
    C_OBJECTS_FILE_NAME,
    C_OBJECTS_PATH_NAME,
)
from cellprofiler_core.preferences import (
    DEFAULT_INPUT_FOLDER_NAME,
    DEFAULT_OUTPUT_FOLDER_NAME,
    NO_FOLDER_NAME,
    ABSOLUTE_FOLDER_NAME,
)

IMAGE_CATEGORIES = (
    C_URL,
    C_FILE_NAME,
    C_PATH_NAME,
)
OBJECTS_CATEGORIES = (
    C_OBJECTS_URL,
    C_OBJECTS_FILE_NAME,
    C_OBJECTS_PATH_NAME,
)
DIR_NONE = "None"
DIR_OTHER = "Elsewhere..."
DIR_ALL = [
    DEFAULT_INPUT_FOLDER_NAME,
    DEFAULT_OUTPUT_FOLDER_NAME,
    NO_FOLDER_NAME,
    ABSOLUTE_FOLDER_NAME,
]
PATH_PADDING = 20



"""
Functions to read and write CellProfiler pipelines.

pipeline._io’s API matches the APIs exposed by the json, marshal, and pickle
standard library modules.
"""

from ._v5 import dump as dump_v5
from ._v6 import dump as dump_v6


def dump(pipeline, fp, save_image_plane_details=True, version=5, sanitize=False):
    if version == 5:
        dump_v5(pipeline, fp, save_image_plane_details, sanitize)
    elif version == 6:
        dump_v6(pipeline, fp, save_image_plane_details)
    else:
        raise NotImplementedError("Invalid pipeline version")



from ._image_name_subscriber_multichoice import ImageNameSubscriberMultiChoice
from ._measurement_multichoice import MeasurementMultiChoice
from ._multichoice import MultiChoice
from ._object_subscriber_multichoice import ObjectSubscriberMultiChoice
from ._subdirectory_filter import SubdirectoryFilter
from ._subscriber_multichoice import SubscriberMultiChoice



import abc
import logging
import sys
import uuid

import docutils.core
import numpy

from ..constants.pipeline import BATCH_STATE
from ..constants.pipeline import MODULE_NAMES
from ..constants.pipeline import MODULE_NOTES
from ..constants.pipeline import MODULE_REVISION_NUMBERS
from ..constants.pipeline import NUMBERS_OF_VARIABLES
from ..constants.pipeline import SETTINGS
from ..constants.pipeline import SHOW_WINDOW
from ..constants.pipeline import VARIABLE_INFO_TYPES
from ..constants.pipeline import VARIABLE_REVISION_NUMBERS
from ..constants.pipeline import VARIABLE_VALUES
from ..setting import HiddenCount
from ..setting import Setting
from ..setting import ValidationError


LOGGER = logging.getLogger(__name__)

class Module:
    """ Derive from the abstract module class to create your own module in Python

    You need to implement the following in the derived class:
    create_settings - create the settings that configure the module.
    settings - return the settings that will be loaded or saved from/to the
               pipeline.
    run - to run the module, producing measurements, etc.

    These methods are optional:
    prepare_settings - adjust the internal state of a module to accept a set of
               stored settings, (e.g., to create the right number of image
               entries in the module's settings.)
    visible_settings - return the settings that will be displayed on the UI
               (default is to use the output of settings())
    upgrade_settings - rewrite a group of settings from a previous version to
               be compatible with the latest version of a module.

    Implement these if you produce measurements:
    get_categories - The category of measurement produced, for instance AreaShape
    get_measurements - The measurements produced by a category
    get_measurement_images - The images measured for a particular measurement
    get_measurement_scales - the scale at which a measurement was taken
    get_measurement_columns - the measurements stored in the database

    The pipeline calls hooks in the module before and after runs and groups.
    The hooks are:
    prepare_run - before run: useful for setting up image sets
    prepare_group - before group: useful for initializing aggregation
    post_group - after group: useful for calculating final aggregation steps
               and for writing out results.
    post_run - use this to perform operations on the results of the experiment,
               for instance on all measurements
    post_pipeline_load - use this to update any settings that require the pipeline
                to be available before they can be adjusted.

    If your module requires state across image_sets, think of storing
    information in the module shared_state dictionary (fetched by
    get_dictionary()).
    """

    def __init__(self):
        if self.__doc__ is None:
            self.__doc__ = sys.modules[self.__module__].__doc__

        self.function = None
        self.__module_num = -1
        self.__settings = []
        self.__notes = []
        self.__show_window = False
        self.__wants_pause = False
        self.__svn_version = "Unknown"
        self.__enabled = True
        self.__variable_revision_number = 0
        self.shared_state = (
            {}
        )  # used for maintaining state between modules, see get_dictionary()
        self.id = uuid.uuid4()
        self.batch_state = numpy.zeros((0,), numpy.uint8)
        # A subclass can override bellow two either by declaring a module_name
        # attribute in the class definition or by assigning to it in
        # the create_settings method.
        # used for citation generation
        if not hasattr(self, "doi"):
            self.doi = []
        # Set the name of the module based on the class name.
        if not hasattr(self, "module_name"):
            self.module_name = self.__class__.__name__
        self.create_settings()

    def to_dict(self) -> dict:
        return {"module_num": self.module_num,
                "notes": self.notes,
                "show_window": self.show_window,
                "wants_pause": self.wants_pause,
                "svn_version": self.svn_version,
                "enabled": self.enabled,
                "variable_revision_number": self.variable_revision_number,
                "batch_state": repr(self.batch_state),
                "module_name": self.__class__.__qualname__,
                "module_path": ".".join([self.__module__, self.__class__.__qualname__])
                }

    def from_dict(self, settings: list, attributes: dict):
        self.module_num = attributes["module_num"]
        self.show_window = attributes["show_window"]
        self.wants_pause = attributes["wants_pause"]
        self.svn_version = attributes["svn_version"]
        self.enabled = attributes["enabled"]
        self.module_path = attributes["module_path"]
        self.module_name = attributes["module_name"]
        setting_values = [setting["value"] for setting in settings]
        if attributes["variable_revision_number"] > self.variable_revision_number:
            LOGGER.warning(f"Loaded module '{self.module_name}' is from a newer version of CellProfiler - "
                            f"v{attributes['variable_revision_number']}, current version is "
                            f"v{self.variable_revision_number}. Settings may load incorrectly.")
        self.set_settings_from_values(
            setting_values, attributes['variable_revision_number'], self.module_path
        )

    @abc.abstractmethod
    def update_settings(self, setting: list):
        pass

    def __setattr__(self, slot, value):
        if hasattr(self, slot) and isinstance(getattr(self, slot), Setting):
            assert isinstance(value, Setting), (
                    "Overwriting %s's %s existing Setting with value of type %s.\nUse __dict__['%s'] = ... to override."
                    % (self.module_name, slot, type(value), slot)
            )
        object.__setattr__(self, slot, value)

    def create_settings(self):
        """Create your settings by subclassing this function

        create_settings is called at the end of initialization.

        You should create the setting variables for your module here:
            # Ask the user for the input image
            self.image_name = cellprofiler_core.settings.ImageNameSubscriber(...)
            # Ask the user for the name of the output image
            self.output_image = cellprofiler_core.settings.ImageNameProvider(...)
            # Ask the user for a parameter
            self.smoothing_size = cellprofiler_core.settings.Float(...)
        """
        pass

    def create_from_handles(self, handles, module_num):
        """Fill a module with the information stored in the handles structure for module # ModuleNum

        Returns a module with the settings decanted from the handles.
        If the revision is old, a different and compatible module can be returned.
        """
        self.__module_num = module_num
        idx = module_num - 1
        settings = handles[SETTINGS][0, 0]
        setting_values = []
        self.__notes = []
        if (
                MODULE_NOTES in settings.dtype.fields
                and settings[MODULE_NOTES].shape[1] > idx
        ):
            n = settings[MODULE_NOTES][0, idx].flatten()
            for x in n:
                if isinstance(x, numpy.ndarray):
                    if len(x) == 0:
                        x = ""
                    else:
                        x = x[0]
                self.__notes.append(x)
        if SHOW_WINDOW in settings.dtype.fields:
            self.__show_window = settings[SHOW_WINDOW][0, idx] != 0
        if BATCH_STATE in settings.dtype.fields:
            # convert from uint8 to array of one string to avoid long
            # arrays, which get truncated by numpy repr()
            self.batch_state = numpy.array(settings[BATCH_STATE][0, idx].tostring())
        setting_count = settings[NUMBERS_OF_VARIABLES][0, idx]
        variable_revision_number = settings[VARIABLE_REVISION_NUMBERS][0, idx]
        module_name = settings[MODULE_NAMES][0, idx][0]
        for i in range(0, setting_count):
            value_cell = settings[VARIABLE_VALUES][idx, i]
            if isinstance(value_cell, numpy.ndarray):
                if numpy.product(value_cell.shape) == 0:
                    setting_values.append("")
                else:
                    setting_values.append(str(value_cell[0]))
            else:
                setting_values.append(value_cell)
        self.set_settings_from_values(
            setting_values, variable_revision_number, module_name
        )

    def prepare_settings(self, setting_values):
        """Do any sort of adjustment to the settings required for the given values

        setting_values - the values for the settings just prior to mapping
                         as done by set_settings_from_values
        This method allows a module to specialize itself according to
        the number of settings and their value. For instance, a module that
        takes a variable number of images or objects can increase or decrease
        the number of relevant settings so they map correctly to the values.

        See cellprofiler_core.modules.measureobjectareashape for an example.
        """
        pass

    def set_settings_from_values(
            self, setting_values, variable_revision_number, module_name
    ):
        """Set the settings in a module, given a list of values

        The default implementation gets all the settings and then
        sets their values using the string passed. A more modern
        module may want to tailor the particular settings set to
        whatever values are in the list or however many values
        are in the list.
        """
        setting_values, variable_revision_number = self.upgrade_settings(
            setting_values, int(variable_revision_number), module_name
        )
        self.prepare_settings(setting_values)
        for v, value in zip(self.settings(), setting_values):
            v.value = value

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):
        """Adjust setting values if they came from a previous revision

        setting_values - a sequence of strings representing the settings
                         for the module as stored in the pipeline
        variable_revision_number - the variable revision number of the
                         module at the time the pipeline was saved. Use this
                         to determine how the incoming setting values map
                         to those of the current module version.
        module_name - the name of the module that did the saving. This can be
                      used to import the settings from another module if
                      that module was merged into the current module

        Overriding modules should return a tuple of setting_values,
        variable_revision_number and True if upgraded to CP 2.0, otherwise
        they should leave things as-is so that the caller can report
        an error.
        """
        return setting_values, variable_revision_number

    def post_pipeline_load(self, pipeline):
        """This is a convenient place to do things to your module after the
           settings have been loaded or initialized"""
        pass

    # https://wiki.python.org/moin/reStructuredText
    @staticmethod
    def _rst_to_html_fragment(source):
        if not isinstance(source, str):
            return ""
        parts = docutils.core.publish_parts(source=source, writer_name="html")

        return parts["body_pre_docinfo"] + parts["fragment"]

    def _get_setting_help(self, setting):
        if isinstance(setting, HiddenCount):
            return ""

        return """\
<div>
    <h4>{SETTING_NAME}</h4>
    <p>{SETTING_DOC}</p>
</div>
        """.format(
            **{
                "SETTING_DOC": self._rst_to_html_fragment(setting.doc),
                "SETTING_NAME": self._rst_to_html_fragment(setting.text),
            }
        )

    def get_help_text(self):
        """Return help text for a module as unformatted strings
        """
        settings_list = self.help_settings()
        if len(settings_list):
            settings_doc = [(setting.text, setting.doc) for setting in settings_list]
        else:
            settings_doc = ""
        mod_doc = self.__doc__ if self.__doc__ else ""
        return mod_doc, settings_doc

    def get_help(self):
        """Return help text for the module

        The default help is taken from your modules docstring and from
        the settings.
        """
        settings_help = (
            """\
<div>
    <h2>Settings:</h2>
    {SETTINGS_DOC}
</div>
""".format(
                **{
                    "SETTINGS_DOC": "\n".join(
                        [
                            self._get_setting_help(setting)
                            for setting in self.help_settings()
                        ]
                    )
                }
            )
            if len(self.help_settings())
            else ""
        )

        return """\
<html style="font-family:arial">
    <body>
        <div>
            {MODULE_DOC}
        </div>
        {SETTINGS_HELP}
    </body>
</html>
""".format(
            **{
                "MODULE_DOC": self._rst_to_html_fragment(self.__doc__),
                "SETTINGS_HELP": settings_help,
                "TITLE": self.module_name,
            }
        )

    def save_to_handles(self, handles):
        from ..setting.text import Name
        from ..setting.subscriber import Subscriber

        module_idx = self.module_num - 1
        setting = handles[SETTINGS][0, 0]
        setting[MODULE_NAMES][0, module_idx] = str(self.module_class())
        setting[MODULE_NOTES][0, module_idx] = numpy.ndarray(
            shape=(len(self.notes), 1), dtype="object"
        )
        for i in range(0, len(self.notes)):
            setting[MODULE_NOTES][0, module_idx][i, 0] = self.notes[i]
        setting[NUMBERS_OF_VARIABLES][0, module_idx] = len(self.settings())
        for i in range(0, len(self.settings())):
            variable = self.settings()[i]
            if len(str(variable)) > 0:
                setting[VARIABLE_VALUES][module_idx, i] = variable.get_unicode_value()
            if isinstance(variable, Name):
                setting[VARIABLE_INFO_TYPES][module_idx, i] = str(
                    "%s indep" % variable.group
                )
            elif isinstance(variable, Subscriber):
                setting[VARIABLE_INFO_TYPES][module_idx, i] = str(variable.group)
        setting[VARIABLE_REVISION_NUMBERS][
            0, module_idx
        ] = self.variable_revision_number
        setting[MODULE_REVISION_NUMBERS][0, module_idx] = 0
        setting[SHOW_WINDOW][0, module_idx] = 1 if self.show_window else 0
        # convert from single-element array with a long string to an
        # array of uint8, to avoid string encoding isues in .MAT
        # format.
        setting[BATCH_STATE][0, module_idx] = numpy.fromstring(
            self.batch_state.tostring(), numpy.uint8
        )

    @staticmethod
    def in_batch_mode():
        """Return True if the module knows that the pipeline is in batch mode"""
        return None

    def change_causes_prepare_run(self, setting):
        """Check to see if changing the given setting means you have to restart

        Some settings, esp in modules like LoadImages, affect more than
        the current image set when changed. For instance, if you change
        the name specification for files, you have to reload your image_set_list.
        Override this and return True if changing the given setting means
        that you'll have to call "prepare_run".
        """
        return False

    def turn_off_batch_mode(self):
        """Reset the module to an editable state if batch mode is on

        A module is allowed to create hidden information that it uses
        to turn batch mode on or to save state to be used in batch mode.
        This call signals that the pipeline has been opened for editing,
        even if it is a batch pipeline; all modules should be restored
        to a state that's appropriate for creating a batch file, not
        for running a batch file
        """
        pass

    def test_valid(self, pipeline):
        """Test to see if the module is in a valid state to run

        Throw a ValidationError exception with an explanation if a module is not valid.
        """
        try:
            for setting in self.visible_settings():
                setting.test_valid(pipeline)
            self.validate_module(pipeline)
        except ValidationError as ve:
            raise ve
        except Exception as e:
            raise ValidationError(
                "Exception in cpmodule.test_valid %s" % e, self.visible_settings()[0]
            )

    def test_module_warnings(self, pipeline):
        """Test to see if there are any troublesome setting values in the module

        Throw a ValidationError exception with an explanation if a module
        is likely to be misconfigured. An example is if ExportToDatabase is
        not the last module.
        """
        try:
            for setting in self.visible_settings():
                setting.test_setting_warnings(pipeline)
            self.validate_module_warnings(pipeline)
        except ValidationError as instance:
            raise instance
        except Exception as e:
            raise ValidationError(
                "Exception in cpmodule.test_valid %s" % e, self.visible_settings()[0]
            )

    def validate_module(self, pipeline):
        """Implement this to validate module settings

        Module implementers should implement validate_module to
        further validate a module's settings. For instance, load_data
        checks the .csv file that it uses in validate_module to ensure
        that the user has chosen a valid .csv file.

        Throw a cps.ValidationError, selecting the most egregiously offending
        setting to indicate failure.
        """
        pass

    def validate_module_warnings(self, pipeline):
        """Implement this to flag potentially dangerous settings

        Module implementers should implement validate_module_warnings to
        find setting combinations that can cause unexpected results.
        Implementers should throw a cps.ValidationError, selecting the
        most egregiously offending setting to indicate failure.
        """
        pass

    def other_providers(self, group):
        """Return a list of hidden name/object/etc. providers supplied by the module for this group

        group - a group supported by a subclass of NameProvider

        This routine returns additional providers beyond those that
        are listed by the module's visible_settings.
        """
        return []

    def get_module_num(self):
        """Get the module's index number

        The module's index number or ModuleNum is a one-based index of its
        execution position in the pipeline. It can be used to predict what
        modules have been run (creating whatever images and measurements
        those modules create) previous to a given module.
        """
        return self.__module_num

    def set_module_num(self, module_num):
        """Change the module's one-based index number in the pipeline

        """
        self.__module_num = module_num

    module_num = property(get_module_num, set_module_num)

    def module_class(self):
        """The class to instantiate, except for the special case of matlab modules.

        """
        return self.__module__ + "." + self.module_name

    def get_enabled(self):
        """True if the module should be executed, False if it should be ignored.

        """
        return self.__enabled

    def set_enabled(self, enable):
        self.__enabled = enable

    enabled = property(get_enabled, set_enabled)

    def settings(self):
        """Return the settings to be loaded or saved to/from the pipeline

        These are the settings (from cellprofiler_core.settings) that are
        either read from the strings in the pipeline or written out
        to the pipeline. The settings should appear in a consistent
        order so they can be matched to the strings in the pipeline.
        """
        return self.__settings

    def help_settings(self):
        """Override this if you want the settings for help to be in a different order"""
        return self.settings()

    def setting(self, setting_num):
        """Reference a setting by its one-based setting number
        """
        return self.settings()[setting_num - 1]

    def set_settings(self, settings):
        self.__settings = settings

    def visible_settings(self):
        """The settings that are visible in the UI
        """
        return self.settings()

    def get_show_window(self):
        """True if the user wants to see the figure for this module"""
        return self.__show_window

    def set_show_window(self, show_window):
        self.__show_window = show_window

    show_window = property(get_show_window, set_show_window)

    def get_wants_pause(self):
        """True if the user wants to pause at this module while debugging"""
        return self.__wants_pause

    def set_wants_pause(self, wants_pause):
        self.__wants_pause = wants_pause

    wants_pause = property(get_wants_pause, set_wants_pause)

    def get_notes(self):
        """The user-entered notes for a module
        """
        return self.__notes

    def set_notes(self, notes):
        """Give the module new user-entered notes

        """
        if isinstance(notes, str):
            notes = [notes]
        sanitization_dict = {"“": "\"", "”": "\""}
        self.__notes = [''.join(sanitization_dict.get(x,x) for x in note) for note in notes]

    notes = property(get_notes, set_notes)

    def get_svn_version(self):
        return self.__svn_version

    def set_svn_version(self, version):
        self.__svn_version = version

    svn_version = property(get_svn_version, set_svn_version)

    def write_to_handles(self, handles):
        """Write out the module's state to the handles

        """
        pass

    def write_to_text(self, file):
        """Write the module's state, informally, to a text file
        """
        pass

    def prepare_run(self, workspace):
        """Prepare the image set list for a run (& whatever else you want to do)

        workspace - holds the following crucial structures:

            pipeline - the pipeline being run

            module - this module

            measurements - measurements structure that can be populated with
                          a image set file names and metadata.

            image_set_list - add any image sets to the image set list

            frame - parent frame of application if GUI enabled, None if GUI
                    disabled

        return True if operation completed, False if aborted
        """
        return True

    def is_load_module(self):
        """If true, the module will load files and make image sets"""
        return False

    @classmethod
    def is_input_module(cls):
        """If true, the module is one of the input modules

        The input modules are "Images", "Metadata", "NamesAndTypes" and "Groups"
        """
        return False

    @staticmethod
    def is_create_batch_module():
        """If true, the module will pickle the pipeline into a batch file and exit

        This is needed by modules which can't properly operate in a batch
        mode (e.g., do all their work post_run or don't work so well if
        run in parallel)
        """
        return False

    @staticmethod
    def is_aggregation_module():
        """If true, the module uses data from other imagesets in a group

        Aggregation modules perform operations that require access to
        all image sets in a group, generally resulting in an aggregation
        operation during the last image set or in post_group. Examples are
        TrackObjects, MakeProjection and CorrectIllumination_Calculate.
        """
        return False

    def convert(self, pipeline, metadata, namesandtypes, groups):
        """Convert the input processing of this module from the legacy format

        Legacy modules like LoadImages should copy their settings into
        the Metadata, NamesAndTypes and Groups modules when this call is made.

        pipeline - the pipeline being converted

        metadata - the pipeline's Metadata module

        namesandtypes - the pipeline's NamesAndTypes module

        groups - the pipeline's Groups module

        """
        pass

    @staticmethod
    def is_object_identification_module():
        """If true, the module will identify primary, secondary or tertiary objects"""
        return False

    def run(self, workspace):
        """Run the module (abstract method)

        workspace    - The workspace contains
            pipeline     - instance of cpp for this run
            image_set    - the images in the image set being processed
            object_set   - the objects (labeled masks) in this image set
            measurements - the measurements for this run
            frame        - the parent frame to whatever frame is created.
                           None means don't draw.
            display_data - the run() module should store anything to be
                           displayed in this attribute, which will be used in
                           display()

        run() should not attempt to display any data, but should communicate it
        to display() via the workspace.
        """
        pass

    def post_run(self, workspace):
        """Do post-processing after the run completes

        workspace - the workspace at the end of the run
        """
        pass

    def display(self, workspace, figure):
        """Display the results, and possibly intermediate results, as
        appropriate for this module.  This method will be called after
        run() is finished if self.show_window is True.

        The run() method should store whatever data display() needs in
        workspace.display_data.  The module is given a CPFigure to use for
        display in the third argument.
        """
        figure.Close()  # modules that don't override display() shouldn't
        # display anything

    def display_post_group(self, workspace, figure):
        """Display the results of work done post-group

        This method is only called if self.show_window is True

        workspace - the current workspace. workspace.display_data should have
                    whatever information is needed for the display. Numpy arrays
                    lists, tuples, dictionaries and Python builtin objects are
                    allowed.

        figure - the figure to use for the display.
        """
        pass

    def display_post_run(self, workspace, figure):
        """Display results after post_run completes

        workspace - a workspace with pipeline, module and measurements valid

        figure - display results in this CPFigure
        """
        pass

    def prepare_to_create_batch(self, workspace, fn_alter_path):
        """Prepare to create a batch file

        This function is called when CellProfiler is about to create a
        file for batch processing. It gives a module an opportunity to
        change its settings and measurements to adapt to file mount differences
        between the machine that created the pipeline and the machines that
        will run the pipeline. You should implement prepare_to_create_batch
        if your module stores paths in settings or measurements. You should
        call fn_alter_path(path) to update any paths to those of the target
        machine.

        workspace - the workspace including the pipeline, the image_set_list
                    and the measurements that need to be modified.

        fn_alter_path - this is a function that takes a pathname on the local
                        host and returns a pathname on the remote host. It
                        handles issues such as replacing backslashes and
                        mapping mountpoints. It should be called for every
                        pathname stored in the settings or legacy fields.

        Returns True if it succeeds.
        """
        return True

    def get_groupings(self, workspace):
        """Return the image groupings of the image sets in an image set list

        get_groupings is called after prepare_run

        workspace - a workspace with an image_set_list and measurements
                    as prepared by prepare_run.

        returns a tuple of key_names and group_list:
        key_names - the names of the keys that identify the groupings
        group_list - a sequence composed of two-tuples.
                     the first element of the tuple is a dictionary giving
                     the metadata values for the metadata keys
                     the second element of the tuple is a sequence of
                     image numbers comprising the image sets of the group
        For instance, an experiment might have key_names of 'Metadata_Row'
        and 'Metadata_Column' and a group_list of:
        [ ({'Metadata_Row':'A','Metadata_Column':'01'}, [1,97,193]),
          ({'Metadata_Row':'A','Metadata_Column':'02'), [2,98,194]),... ]

        Returns None to indicate that the module does not contribute any
        groupings.
        """
        return None

    def prepare_group(self, workspace, grouping, image_numbers):
        """Prepare to start processing a new grouping

        workspace - the workspace for the group. The pipeline, measurements
                    and image_set_list are valid at this point and you can
                    fill in image_sets at this point.
        grouping - a dictionary that describes the key for the grouping.
                   For instance, { 'Metadata_Row':'A','Metadata_Column':'01'}
        image_numbers - a sequence of the image numbers within the
                   group (image sets can be retreved as
                   image_set_list.get_image_set(image_numbers[i]-1)

        prepare_group is called once after prepare_run if there are no
        groups.
        """
        pass

    def post_group(self, workspace, grouping):
        """Do post-processing after a group completes

        workspace - the workspace at the end of the group
        grouping - the group that's being run
        """
        pass

    def get_measurement_columns(self, pipeline):
        """Return a sequence describing the measurement columns needed by this module

        This call should return one element per image or object measurement
        made by the module during image set analysis. The element itself
        is a 3-tuple:
        first entry: either one of the predefined measurement categories,
                     {"Image", "Experiment" or "Neighbors" or the name of one
                     of the objects.}
        second entry: the measurement name (as would be used in a call
                      to add_measurement)
        third entry: the column data type (for instance, "varchar(255)" or
                     "float")
        """
        return []

    @staticmethod
    def get_object_relationships(pipeline):
        """Return a sequence describing the relationships recorded in measurements

        This method reports the relationships recorded in the measurements
        using add_relate_measurement. Modules that add relationships should
        return one 4-tuple of
        (<relationship-name>, <object-name-1>, <object-name-2>, <when>)
        for every combination of relationship and parent / child objects
        that will be produced during the course of a run.

        <when> is one of cpmeas.MCA_AVAILABILE_EACH_CYCLE or
        cpmeas.MCA_AVAILABLE_POST_GROUP. cpmeas.MCA_AVAILABLE_EACH_CYCLE
        promises that the relationships will be available after each cycle.
        Any relationship with that cycle's image number (as either the
        parent or child) will be inserted if not already present in the database.

        MCA_AVAILABLE_POST_GROUP indicates that the relationship is not available
        until the group has completed - all relationships with a group's
        image number will be written in that case.
        """
        return []

    def get_dictionary(self, ignore=None):
        """Get the dictionary for this module
        """
        return self.shared_state

    def get_dictionary_for_worker(self):
        """Get the dictionary that should be shared between analysis workers

        A module might use the dictionary for cacheing information stored on
        disk or that's difficult to compute. It might also use it to store
        aggregate data, but this data may not be useful to other workers.

        Finally, a module might store Python objects that aren't JSON serializable
        in its dictionary. In these cases, the module should create a dictionary
        that can be JSON serialized in get_dictionary_for_worker and then
        reconstruct the result of JSON deserialization in set_dictionary_in_worker.
        """
        return self.get_dictionary()

    def set_dictionary_for_worker(self, d):
        """Initialize this worker's dictionary using results from first worker

        see get_dictionary_for_worker for details.
        """
        self.get_dictionary().clear()
        self.get_dictionary().update(d)

    def get_categories(self, pipeline, object_name):
        """Return the categories of measurements that this module produces

        object_name - return measurements made on this object (or 'Image' for image measurements)
        """
        return []

    def get_measurements(self, pipeline, object_name, category):
        """Return the measurements that this module produces

        object_name - return measurements made on this object (or 'Image' for image measurements)
        category - return measurements made in this category
        """
        return []

    def get_measurement_images(self, pipeline, object_name, category, measurement):
        """Return a list of image names used as a basis for a particular measure
        """
        return []

    @staticmethod
    def get_measurement_objects(pipeline, object_name, category, measurement):
        """Return a list of secondary object names used as a basis for a particular measure

        object_name - either "Image" or the name of the primary object
        category - the category being measured, for instance "Threshold"
        measurement - the name of the measurement being done

        Some modules output image-wide aggregate measurements in addition to
        object measurements. These must be stored using the "Image" object name
        in order to save a single value. A module can override
        get_measurement_objects to tell the user about the object name in
        those situations.

        In addition, some modules may make use of two segmentations, for instance
        when measuring the total value of secondary objects related to primary
        ones. This mechanism can be used to identify the secondary objects used.
        """
        return []

    def get_measurement_scales(
            self, pipeline, object_name, category, measurement, image_name
    ):
        """Return a list of scales (eg for texture) at which a measurement was taken
        """
        return []

    def is_image_from_file(self, image_name):
        """Return True if this module loads this image name from a file."""
        from ..setting.subscriber import FileImageSubscriber

        for setting in self.settings():
            if (
                    isinstance(setting, FileImageSubscriber, )
                    and setting.value == image_name
            ):
                return True
        return False

    @staticmethod
    def should_stop_writing_measurements():
        """Returns True if measurements should not be taken after this module

        The ExportToDatabase and ExportToExcel modules expect that no
        measurements will be recorded in latter modules. This function
        returns False in the default, indicating that measurements should
        keep being made, but returns True for these modules, indicating
        that any subsequent modules will lose their measurements and should
        not write any.
        """
        return False

    def needs_default_image_folder(self, pipeline):
        """Returns True if the module needs the default image folder

        pipeline - pipeline being run

        Legacy modules might need the default image folder as does any module
        that uses the DirectoryPath setting.
        """
        from ..setting.text import Directory

        for setting in self.visible_settings():
            if isinstance(setting, Directory):
                return True
        return False

    def obfuscate(self):
        """Erase any sensitive information in a module's settings

        You should implement "obfuscate" to erase information like
        passwords or file names so that the pipeline can be uploaded
        for error reporting without revealing that information.
        """
        pass

    def on_activated(self, workspace):
        """Called when the module is activated in the GUI

        workspace - the workspace that's currently running

        on_activated is here to give modules the chance to modify other
        elements of the pipeline, such as the image plane details or image
        set list. You're allowed to modify these parts of the pipeline
        in the UI thread until on_deactivated is called.
        """
        pass

    def on_deactivated(self):
        """Called when the module is deactivated in the GUI

        This is the signal that the settings have been unhooked from the
        GUI and can't be used to edit the pipeline
        """
        pass

    def on_setting_changed(self, setting, pipeline):
        """Called when a setting has been changed in the GUI"""
        pass

    def volumetric(self):
        return False



from ._event import Event


class ModuleEnabled(Event):
    """A module was enabled

    module - the module that was enabled.
    """

    def __init__(self, module):
        """Constructor

        module - the module that was enabled
        """
        super(self.__class__, self).__init__(is_pipeline_modification=True)
        self.module = module

    def event_type(self):
        return "Module enabled"



import logging
import os
import re
import sys

from ....module import Module
from ....constants.modules import (
    do_not_override,
    should_override,
    all_modules,
    svn_revisions,
    pymodules,
    badmodules,
    builtin_modules,
    renamed_modules,
    unimplemented_modules,
    depricated_modules,
    replaced_modules,
)
from ..plugins import load_plugins


LOGGER = logging.getLogger(__name__)

def check_module(module, name):
    if hasattr(module, "do_not_check"):
        return
    assert (
        name == module.module_name
    ), "Module %s should have module_name %s (is %s)" % (name, name, module.module_name)
    for method_name in do_not_override:
        assert getattr(module, method_name) == getattr(
            Module, method_name
        ), "Module %s should not override method %s" % (name, method_name)
    for method_name in should_override:
        assert getattr(module, method_name) != getattr(
            Module, method_name
        ), "Module %s should override method %s" % (name, method_name)


def find_cpmodule(m):
    """Returns the CPModule from within the loaded Python module

    m - an imported module

    returns the CPModule class
    """
    modulename = os.path.splitext(os.path.basename(m.__file__))[0]
    for key, item in list(m.__dict__.items()):
        if key.lower() == modulename and issubclass(item, Module):
            return item
    raise ValueError(
        "Could not find cellprofiler_core.module.Module class in %s" % m.__file__
    )


def fill_modules():
    del pymodules[:]
    del badmodules[:]
    all_modules.clear()
    svn_revisions.clear()

    def add_module(mod, check_svn, class_name=None):
        try:
            m = __import__(mod, globals(), locals(), ["__all__"], 0)
            if class_name:
                cp_module = m.__dict__[class_name]
                assert issubclass(cp_module, Module)
            else:
                cp_module = find_cpmodule(m)
            name = cp_module.module_name
        except Exception as e:
            LOGGER.warning("Could not load %s", mod, exc_info=True)
            badmodules.append((mod, e))
            return

        try:
            pymodules.append(m)
            if name in all_modules:
                LOGGER.warning(
                    "Multiple definitions of module %s\n\told in %s\n\tnew in %s",
                    name,
                    sys.modules[all_modules[name].__module__].__file__,
                    m.__file__,
                )
            all_modules[name] = cp_module
            check_module(cp_module, name)
            # attempt to instantiate
            if not hasattr(cp_module, "do_not_check"):
                cp_module()
            if check_svn and hasattr(m, "__version__"):
                match = re.match("^\$Revision: ([0-9]+) \$$", m.__version__)
                if match is not None:
                    svn_revisions[name] = match.groups()[0]
        except Exception as e:
            LOGGER.warning("Failed to load %s", name, exc_info=True)
            badmodules.append((mod, e))
            if name in all_modules:
                del all_modules[name]
                del pymodules[-1]

    # Import core modules
    for modname, classname in builtin_modules.items():
        add_module("cellprofiler_core.modules." + modname, True, class_name=classname)

    # Import CellProfiler modules if CellProfiler is installed
    cpinstalled = False
    try:
        import cellprofiler.modules

        cpinstalled = True
    except ImportError:
        print("No CellProfiler installation detected, only base modules will be loaded")
    if cpinstalled:
        for modname, classname in cellprofiler.modules.builtin_modules.items():
            add_module("cellprofiler.modules." + modname, True, class_name=classname)

    if len(badmodules) > 0:
        LOGGER.warning(
            "could not load these modules: %s", ",".join([x[0] for x in badmodules])
        )


def add_module_for_tst(module_class):
    all_modules[module_class.module_name] = module_class


def get_module_class(module_name):
    module_class = module_name.split(".")[-1]
    if module_class not in all_modules:
        if module_class in renamed_modules:
            return all_modules[renamed_modules[module_class]]
        if module_class in unimplemented_modules:
            raise ValueError(
                (
                    "The %s module has not yet been implemented. "
                    "It will be available in a later version "
                    "of CellProfiler."
                )
                % module_class
            )
        if module_class in depricated_modules:
            raise ValueError(
                (
                    "The %s module has been deprecated and will "
                    "not be implemented in CellProfiler 4.0."
                )
                % module_class
            )
        if module_class in replaced_modules:
            raise ValueError(
                (
                    "The %s module no longer exists. You can find "
                    "similar functionality in: %s"
                )
                % (module_class, ", ".join(replaced_modules[module_class]))
            )
        raise ValueError("Could not find the %s module" % module_class)
    return all_modules[module_class]


def instantiate_module(module_name):
    module = get_module_class(module_name)()
    if module_name in svn_revisions:
        module.svn_version = svn_revisions[module_name]
    return module


def get_module_names():
    names = list(all_modules.keys())
    names.sort()
    return names


def reload_modules():
    for m in pymodules:
        try:
            del sys.modules[m.__name__]
        except:
            pass
    fill_modules()
    load_plugins(modules_only=True)



from ._file_image import FileImage


class MovieFrameImage(FileImage):
    """Provide an image by filename:frame, loading the file as it is requested
    """

    def __init__(self, name, pathname, filename, frame, rescale_range=None):
        super(MovieFrameImage, self).__init__(
            name,
            pathname,
            filename,
            rescale_range=rescale_range,
            metadata_rescale=False,
            index=frame
        )



import pathlib
import urllib
import logging

from cellprofiler_core.constants.image import FILE_SCHEME, PASSTHROUGH_SCHEMES
from cellprofiler_core.utilities.image import is_file_url

LOGGER = logging.getLogger(__name__)

def pathname2url(path):
    """Convert the unicode path to a file: url"""
    lower_path = path.lower()
    if any((lower_path.startswith(x) for x in PASSTHROUGH_SCHEMES)):
        return path
    path_object = pathlib.Path(path)
    if path_object.is_absolute():
        # Produces a valid URI regardless of platform.
        return path_object.as_uri()
    # Produces CellProfiler's interpretation of a relative path URI.
    return FILE_SCHEME + urllib.request.pathname2url(path)

def url2pathname(url):
    lower_url = url.lower()

    if any((lower_url.startswith(x) for x in PASSTHROUGH_SCHEMES)):
        return url

    if is_file_url(url):
        return urllib.request.url2pathname(url[len(FILE_SCHEME):])



import h5py

from ..hdf5_dict import HDF5FileList
from ..hdf5_dict import HDF5Dict


def is_workspace_file(path):
    """Return True if the file along the given path is a workspace file"""
    if not h5py.is_hdf5(path):
        return False
    h5file = h5py.File(path, mode="r")
    try:
        if not HDF5FileList.has_file_list(h5file):
            return False
        return HDF5Dict.has_hdf5_dict(h5file)
    finally:
        h5file.close()



import os
import sys
import traceback

from ._reader import Reader
from ..constants.reader import ALL_READERS, builtin_readers, BAD_READERS, AVAILABLE_READERS

import logging

from ..preferences import config_read_typed

LOGGER = logging.getLogger(__name__)


def add_reader(reader_name, check_svn, class_name=None):
    LOGGER.debug(f"Registering {reader_name}")
    try:
        rdr = __import__(reader_name, globals(), locals(), ["__all__"], 0)
        if class_name:
            cp_reader = rdr.__dict__[class_name]
            assert issubclass(cp_reader, Reader)
        else:
            cp_reader = find_cp_reader(rdr)
        name = cp_reader.reader_name
    except Exception as e:
        reader_name = reader_name.replace("cellprofiler_core.readers.", "")
        LOGGER.warning(f"Could not load {reader_name}", exc_info=True)
        BAD_READERS[reader_name] = traceback.format_exc()
        return
    try:
        if name in ALL_READERS:
            LOGGER.warning(
                "Multiple definitions of reader %s\n\told in %s\n\tnew in %s",
                name,
                sys.modules[ALL_READERS[name].__module__].__file__,
                rdr.__file__,
            )
        ALL_READERS[name] = cp_reader
    except Exception as e:
        LOGGER.warning("Failed to load %s", name, exc_info=True)
        BAD_READERS[reader_name] = traceback.format_exc()
        if name in ALL_READERS:
            del ALL_READERS[name]


def fill_readers(check_config=True):
    ALL_READERS.clear()
    BAD_READERS.clear()

    # Import core modules
    for reader_name, classname in builtin_readers.items():
        add_reader("cellprofiler_core.readers." + reader_name, True, class_name=classname)

    if len(BAD_READERS) > 0:
        LOGGER.warning(
            f"could not load these readers: {', '.join(BAD_READERS)}"
        )

    activate_readers(check_config)


def activate_readers(check_config=True):
    AVAILABLE_READERS.clear()
    for reader_name, classname in ALL_READERS.items():
        if check_config:
            enabled = config_read_typed(f'Reader.{reader_name}.enabled', bool)
            if enabled or enabled is None:
                AVAILABLE_READERS[reader_name] = classname
        else:
            AVAILABLE_READERS[reader_name] = classname

    if len(AVAILABLE_READERS) == 0:
        LOGGER.critical("No image readers are available, CellProfiler won't be able to load data!")
    else:
        LOGGER.debug("Image readers available and active: %s", ", ".join(AVAILABLE_READERS))

def filter_active_readers(readers_to_keep, from_all_readers=True, by_module_name=True):
    """
    Filter out readers that are not in the readers_to_keep list.
    from_all_readers - if True, use ALL_READERS instead of AVAILABLE_READERS
    by_module_name - if True, use the module name, if False use the reader name
    """
    if from_all_readers:
        readers_to_filter = ALL_READERS.copy()
    else:
        readers_to_filter = AVAILABLE_READERS.copy()

    if by_module_name:
        filtered_readers = {reader_name: classname for reader_name, classname in readers_to_filter.items() if classname.__module__.split('.')[-1] in readers_to_keep}
    else:
        filtered_readers = {reader_name: classname for reader_name, classname in readers_to_filter.items() if classname.reader_name in readers_to_keep}

    if len(filtered_readers) == 0:
        LOGGER.critical("No image readers are available after filtering, CellProfiler won't be able to load data!")
    else:
        LOGGER.debug("Image readers available and active after filtering: %s", ", ".join(filtered_readers))

    AVAILABLE_READERS.clear()
    AVAILABLE_READERS.update(filtered_readers)

def find_cp_reader(rdr):
    """Returns the CPModule from within the loaded Python module

    m - an imported module

    returns the CPModule class
    """
    reader_name = os.path.splitext(os.path.basename(rdr.__file__))[0]
    for key, item in list(rdr.__dict__.items()):
        if key.lower() == reader_name and issubclass(item, Reader):
            return item
    raise ValueError(
        "Could not find cellprofiler_core.reader.Reader class in %s" % rdr.__file__
    )


def get_image_reader_class(image_file, use_cached_name=True, volume=False):
    if not AVAILABLE_READERS:
        raise Exception("No image readers are enabled.\n"
                        "Please check reader configuration in the File menu.")
    if use_cached_name and image_file.preferred_reader in AVAILABLE_READERS:
        reader_class = get_image_reader_by_name(image_file.preferred_reader)
        return reader_class
    LOGGER.debug(f"Choosing reader for {image_file.filename}")
    best_reader = None
    best_value = 5
    for reader_name, reader_class in AVAILABLE_READERS.items():
        result = reader_class.supports_format(image_file, volume=volume, allow_open=False)
        if result == 1:
            LOGGER.debug(f"Selected {reader_name}")
            image_file.preferred_reader = reader_name
            return reader_class
        elif 1 < result < best_value:
            best_value = result
            best_reader = reader_class
    if best_reader is None:
        raise NotImplementedError(f"No reader available for {image_file.filename}")
    LOGGER.debug(f"Selected {best_reader}")
    return best_reader


def get_image_reader(image_file, use_cached_name=True, volume=False):
    reader_class = get_image_reader_class(image_file, use_cached_name=use_cached_name, volume=volume)
    return reader_class(image_file)


def get_image_reader_by_name(reader_name):
    assert reader_name in ALL_READERS, f"Reader {reader_name} not found"
    if reader_name not in AVAILABLE_READERS:
        LOGGER.warning(f"Requested reader {reader_name} which is disabled by config."
                       f"CellProfiler will use this reader anyway.")
    return ALL_READERS[reader_name]



import cellprofiler_core.utilities.legacy
from ._setting import Setting


class Table(Setting):
    """The Table setting displays a table of values"""

    ATTR_ERROR = "Error"

    def __init__(
        self,
        text,
        min_size=(400, 300),
        max_field_size=30,
        use_sash=False,
        corner_button=None,
        **kwargs,
    ):
        """Constructor

        text - text label to display to the left of the table
        min_size - initial size of the table before user stretches it
        max_field_size - any field with more than this # of characters will
                         be truncated using an ellipsis.
        use_sash - if True, place the table in the bottom resizable sash.
                   if False, place the table inline
        corner_button - if defined, consists of keyword arguments for the corner
                        button mixin: dict(fn_clicked=<function>, label=<label>,
                        tooltip=<tooltip>)
        """
        super(self.__class__, self).__init__(text, "", **kwargs)
        self.column_names = []
        self.data = []
        self.row_attributes = {}
        self.cell_attributes = {}
        self.min_size = min_size
        self.max_field_size = max_field_size
        self.use_sash = use_sash
        self.corner_button = corner_button

    def insert_column(self, index, column_name):
        """Insert a column at the given index

        index - the zero-based index of the column's position

        column_name - the name of the column

        Adds the column to the table and sets the value for any existing
        rows to None.
        """
        self.column_names.insert(index, column_name)
        for row in self.data:
            row.insert(index, None)

    def add_rows(self, columns, data):
        """Add rows to the table

        columns - define the columns for each row of data

        data - rows of data to add. Each field in a row is placed
               at the column indicated by "columns"
        """
        indices = [
            columns.index(c) if c in columns else None for c in self.column_names
        ]
        for row in data:
            self.data.append(
                [None if index is None else row[index] for index in indices]
            )

    def sort_rows(self, columns):
        """Sort rows based on values in columns"""
        indices = [self.column_names.index(c) for c in columns]

        def compare_fn(row1, row2):
            for index in indices:
                x = cellprofiler_core.utilities.legacy.cmp(row1[index], row2[index])
                if x != 0:
                    return x
            return 0

        self.data.sort(compare_fn)

    def clear_rows(self):
        self.data = []
        self.row_attributes = {}
        self.cell_attributes = {}

    def clear_columns(self):
        self.column_names = []

    def get_data(self, row_index, columns):
        """Get the column values for a given row or rows

        row_index - can either be the index of one row or can be a slice or list
                    of rows

        columns - the names of the columns to fetch, in the order they will
                  appear in the row
        """
        column_indices = [self.column_names.index(c) for c in columns]
        if isinstance(row_index, int):
            row_index = slice(row_index, row_index + 1)
        return [[row[ci] for ci in column_indices] for row in self.data[row_index]]

    def set_row_attribute(self, row_index, attribute, set_attribute=True):
        """Set an attribute on a row

        row_index - index of row in question

        attribute - one of the ATTR_ values, for instance ATTR_ERROR

        set_attribute - True to set, False to clear
        """
        if set_attribute:
            if row_index in self.row_attributes:
                self.row_attributes[row_index].add(attribute)
            else:
                self.row_attributes[row_index] = {attribute}
        else:
            if row_index in self.row_attributes:
                s = self.row_attributes[row_index]
                s.remove(attribute)
                if len(s) == 0:
                    del self.row_attributes[row_index]

    def get_row_attributes(self, row_index):
        """Get the set of attributes on a row

        row_index - index of the row being queried

        returns None if no attributes or a set of attributes set on the row
        """
        return self.row_attributes.get(row_index, None)

    def set_cell_attribute(self, row_index, column_name, attribute, set_attribute=True):
        """Set an attribute on a cell

        row_index - index of row in question

        column_name - name of the cell's column

        attribute - one of the ATTR_ values, for instance ATTR_ERROR

        set_attribute - True to set, False to clear
        """
        key = (row_index, self.column_names.index(column_name))
        if set_attribute:
            if key in self.cell_attributes:
                self.cell_attributes[key].add(attribute)
            else:
                self.cell_attributes[key] = {attribute}
        else:
            if key in self.cell_attributes:
                s = self.cell_attributes[key]
                s.remove(attribute)
                if len(s) == 0:
                    del self.cell_attributes[key]

    def get_cell_attributes(self, row_index, column_name):
        """Get the set of attributes on a row

        row_index - index of the row being queried

        returns None if no attributes or a set of attributes set on the row
        """
        key = (row_index, self.column_names.index(column_name))
        return self.cell_attributes.get(key, None)



class Grid:
    def __init__(self):
        self.x_location_of_lowest_x_spot = None
        self.y_location_of_lowest_y_spot = None
        self.x_spacing = None
        self.y_spacing = None
        self.rows = None
        self.columns = None
        self.vert_lines_x = None
        self.vert_lines_y = None
        self.horiz_lines_x = None
        self.horiz_lines_y = None
        self.spot_table = None
        self.total_height = None
        self.total_width = None
        self.y_locations = None
        self.x_locations = None
        self.left_to_right = None
        self.top_to_bottom = None
        self.image_width = None
        self.image_height = None

    def serialize(self):
        return dict(
            (k, v) for k, v in list(self.__dict__.items()) if not k.startswith("_")
        )

    def deserialize(self, serialized_info):
        self.__dict__.update(serialized_info)



import logging
import queue
import threading

import zmq

import cellprofiler_core.utilities.zmq
from cellprofiler_core.utilities.zmq._analysis_context import AnalysisContext
from .communicable import Communicable
from .communicable.reply.upstream_exit import BoundaryExited
from .communicable.request import AnalysisRequest, Request
from ...constants.worker import NOTIFY_STOP, NOTIFY_RUN


LOGGER = logging.getLogger(__name__)

class Boundary:
    """This object serves as the interface between a ZMQ socket passing
    Requests and Replies, and a thread or threads serving those requests.
    Received requests are received on a ZMQ socket and placed on upward_queue,
    and notify_all() is called on updward_cv.  Replies (via the Request.reply()
    method) are dispatched to their requesters via a downward queue.

    The Boundary wakes up the socket thread via the notify socket. This lets
    the socket thread poll for changes on the notify and request sockets, but
    allows it to receive Python objects via the downward queue.
    """

    def __init__(self, zmq_address, port=None):
        """Construction

        zmq_address - the address for announcements and requests
        port - the port for announcements, defaults to random
        """
        self.analysis_context = None
        self.analysis_context_lock = threading.RLock()
        #
        # Dictionary of request dictionary to queue for handler
        # (not including AnalysisRequest)
        #
        self.request_dictionary = {}
        self.zmq_context = zmq.Context()
        # Set linger to 0 so that all sockets close without
        # waiting for transmission during shutdown.
        self.zmq_context.setsockopt(zmq.LINGER, 0)
        self.zmq_address = zmq_address
        # The downward queue is used to feed replies to the socket thread
        self.downward_queue = queue.Queue()

        # Not entirely convinced that threadlocal is still needed, but meh.
        self.threadlocal = (
            threading.local()
        )  # for connecting to notification socket, and receiving replies

        #
        # Socket for telling workers whether to keep running
        # We should ONLY send over this socket from the spin thread
        # We need to create it here so that we can transmit the address to
        # workers on creation.
        #
        self.keepalive_socket = self.zmq_context.socket(zmq.PUB)

        self.keepalive_socket_port = self.keepalive_socket.bind_to_random_port(
            "tcp://127.0.0.1"
        )
        self.keepalive_address = f"tcp://127.0.0.1:{self.keepalive_socket_port}"

        self.thread = threading.Thread(
            target=self.spin,
            name="Boundary spin() thread",
        )
        self.thread.start()

    """Notify a request class handler of a request"""
    NOTIFY_REQUEST = "request"
    """Notify the socket thread that a reply is ready to be sent"""
    NOTIFY_REPLY_READY = "reply ready"
    """Cancel an analysis. The analysis ID is the second part of the message"""
    NOTIFY_CANCEL_ANALYSIS = "cancel analysis"
    """Stop the socket thread"""
    NOTIFY_STOP = "stop"

    def register_analysis(self, analysis_id, upward_queue):
        """Register a queue to receive analysis requests

        analysis_id - the analysis ID embedded in each analysis request

        upward_queue - place the requests on this queue
        """
        with self.analysis_context_lock:
            self.analysis_context = AnalysisContext(
                analysis_id, upward_queue, self.analysis_context_lock
            )
        LOGGER.debug(f"Registered analysis as id {analysis_id}")

    def register_request_class(self, cls_request, upward_queue):
        """Register a queue to receive requests of the given class

        cls_request - requests that match isinstance(request, cls_request) will
                      be routed to the upward_queue

        upward_queue - queue that will receive the requests
        """
        self.request_dictionary[cls_request] = upward_queue

    def enqueue_reply(self, req, rep):
        """Enqueue a reply to be sent from the boundary thread

        req - original request
        rep - the reply to the request
        """
        self.send_to_boundary_thread(self.NOTIFY_REPLY_READY, (req, rep))

    def cancel(self, analysis_id):
        """Cancel an analysis

        All requests with the given analysis ID will get a BoundaryExited
        reply after this call returns.
        """
        with self.analysis_context_lock:
            if self.analysis_context.cancelled:
                return
            self.analysis_context.cancel()
        response_queue = queue.Queue()
        self.send_to_boundary_thread(
            self.NOTIFY_CANCEL_ANALYSIS, (analysis_id, response_queue)
        )
        response_queue.get()

    def handle_cancel(self, analysis_id, response_queue):
        """Handle cancellation in the boundary thread"""
        with self.analysis_context_lock:
            self.analysis_context.handle_cancel()
        response_queue.put("OK")

    def join(self):
        """Join to the boundary thread.

        Note that this should only be done at a point where no worker truly
        expects a reply to its requests.
        """
        self.send_to_boundary_thread(self.NOTIFY_STOP, None)
        self.thread.join()

    def spin(self):
        try:
            # socket for handling downward notifications
            selfnotify_socket = self.zmq_context.socket(zmq.SUB)
            selfnotify_socket.connect(
                cellprofiler_core.utilities.zmq.NOTIFY_SOCKET_ADDR)
            selfnotify_socket.setsockopt(zmq.SUBSCRIBE, b"")

            # socket where we receive Requests
            request_socket = self.zmq_context.socket(zmq.ROUTER)
            request_socket.setsockopt(zmq.LINGER, 0)
            request_socket.set_hwm(2000)
            request_port = request_socket.bind_to_random_port(
                self.zmq_address)
            self.request_address = self.zmq_address + (":%d" % request_port)

            poller = zmq.Poller()
            poller.register(selfnotify_socket, zmq.POLLIN)
            poller.register(request_socket, zmq.POLLIN)

            received_stop = False
            while not received_stop:
                # Tell the workers to stay alive
                self.keepalive_socket.send(NOTIFY_RUN)
                poll_result = poller.poll(1000)
                #
                # Under all circumstances, read everything from the queue
                #
                try:
                    while True:
                        notification, arg = self.downward_queue.get_nowait()
                        if notification == self.NOTIFY_REPLY_READY:
                            req, rep = arg
                            self.handle_reply(req, rep)
                        elif notification == self.NOTIFY_CANCEL_ANALYSIS:
                            analysis_id, response_queue = arg
                            self.handle_cancel(analysis_id, response_queue)
                        elif notification == self.NOTIFY_STOP:
                            received_stop = True
                except queue.Empty:
                    pass
                #
                # Then process the poll result
                #
                for s, state in poll_result:
                    if s == selfnotify_socket:
                        # This isn't really used for message transmission.
                        # Instead, it pokes this spin thread to get it to check
                        # the non-ZMQ message queue.
                        msg = selfnotify_socket.recv()
                        # Let's watch out for stop signals anyway
                        if msg == NOTIFY_STOP:
                            LOGGER.warning("Captured a stop message over zmq")
                            received_stop = True
                        continue
                    req = Communicable.recv(s, routed=True)
                    req.set_boundary(self)
                    if not isinstance(req, AnalysisRequest):
                        for request_class in self.request_dictionary:
                            if isinstance(req, request_class):
                                q = self.request_dictionary[request_class]
                                q.put([self, self.NOTIFY_REQUEST, req])
                                break
                        else:
                            LOGGER.warning(
                                "Received a request that wasn't an AnalysisRequest: %s"
                                % str(type(req))
                            )
                            req.reply(BoundaryExited())
                        continue
                    if s != request_socket:
                        # Request is on the external socket
                        LOGGER.warning("Received a request on the external socket")
                        req.reply(BoundaryExited())
                        continue
                    #
                    # Filter out requests for cancelled analyses.
                    #
                    with self.analysis_context_lock:
                        if not self.analysis_context.enqueue(req):
                            continue
            LOGGER.debug("boundary received stop")
            # Give the workers an explicit stop command
            self.keepalive_socket.send(NOTIFY_STOP)
            self.keepalive_socket.close()
            #
            # We assume here that workers trying to communicate with us will
            # be shut down without needing replies to pending requests.
            # There's not much we can do in terms of handling that in a more
            # orderly fashion since workers might be formulating requests as or
            # after we have shut down. But calling cancel on all the analysis
            # contexts will raise exceptions in threads waiting for a rep/rep.
            #
            # You could call analysis_context.handle_cancel() here, what if it
            # blocks?
            with self.analysis_context_lock:
                self.analysis_context.cancel()
                request_dict_values = self.request_dictionary.values()
                LOGGER.debug(f"joining {len(request_dict_values)} threads")
                for request_class_queue in request_dict_values:
                    #
                    # Tell each response class to stop. Wait for a reply
                    # which may be a thread instance. If so, join to the
                    # thread so there will be an orderly shutdown.
                    #
                    response_queue = queue.Queue()
                    request_class_queue.put(
                        [self, self.NOTIFY_STOP, response_queue]
                    )
                    thread = response_queue.get()
                    LOGGER.debug("joining on thread from response_queue")
                    if isinstance(thread, threading.Thread):
                        thread.join()

            request_socket.close()
            selfnotify_socket.close()
            LOGGER.debug("Shutting down the boundary thread")
        except:
            #
            # Pretty bad - a logic error or something extremely unexpected
            #              We're close to hosed here, better die an ugly death.
            #
            LOGGER.critical("Unhandled exception in boundary thread.",
                             exc_info=True)
            import os
            os._exit(-1)

    def send_to_boundary_thread(self, msg, arg):
        """Send a message to the boundary thread via the notify socket

        Send a wakeup call to the boundary thread by sending arbitrary
        data to the notify socket, placing the real objects of interest
        on the downward queue.

        msg - message placed in the downward queue indicating the purpose
              of the wakeup call

        args - supplementary arguments passed to the boundary thread via
               the downward queue.
        """
        if not hasattr(self.threadlocal, "notify_socket"):
            self.threadlocal.notify_socket = self.zmq_context.socket(zmq.PUB)
            self.threadlocal.notify_socket.setsockopt(zmq.LINGER, 0)
            self.threadlocal.notify_socket.connect(
                cellprofiler_core.utilities.zmq.NOTIFY_SOCKET_ADDR
            )
        self.downward_queue.put((msg, arg))
        self.threadlocal.notify_socket.send(b"WAKE UP!")

    def handle_reply(self, req, rep):
        if not isinstance(req, AnalysisRequest):
            assert isinstance(req, Request)
            Communicable.reply(req, rep)
            return

        with self.analysis_context_lock:
            self.analysis_context.reply(req, rep)



class DisplayPostRun:
    """Request a post-run display

    This is a message sent to the UI from the analysis worker"""

    def __init__(self, module_num, display_data):
        self.module_num = module_num
        self.display_data = display_data



from ._abstract_image import AbstractImage


class VanillaImage(AbstractImage):
    """This image provider returns the image given to it in the constructor

    """

    def __init__(self, name, image):
        """Constructor takes the name of the image and the CellProfiler.Image.Image instance to be returned
        """
        self.__name = name
        self.__image = image

    def get_name(self):
        return self.__name

    def provide_image(self, image_set):
        return self.__image

    def release_memory(self):
        self.__image = None



from ._image_set_success import ImageSetSuccess


class ImageSetSuccessWithDictionary(ImageSetSuccess):
    def __init__(self, analysis_id, image_set_number, shared_dicts):
        ImageSetSuccess.__init__(self, analysis_id, image_set_number=image_set_number)

        self.shared_dicts = shared_dicts



import os.path

from .._text import Text
from ..._validation_error import ValidationError


class Pathname(Text):
    """A setting that displays a path name

    text - text to display to right
    value - initial value
    wildcard - wildcard to filter files in browse dialog
    """

    def __init__(self, text, value="", *args, **kwargs):
        kwargs = kwargs.copy()
        if "wildcard" in kwargs:
            self.wildcard = kwargs["wildcard"]
            del kwargs["wildcard"]
        else:
            self.wildcard = "All files (*.*)|*.*"
        super(Pathname, self).__init__(text, value, *args, **kwargs)

    def test_valid(self, pipeline):
        if not os.path.isfile(self.value):
            raise ValidationError("Can't find file, %s" % self.value, self)

    def alter_for_create_batch(self, fn_alter):
        self.value = fn_alter(self.value)



import os

from ._does_not_predicate import DoesNotPredicate
from ._does_predicate import DoesPredicate
from ._file_predicate import FilePredicate
from ._filter import CONTAINS_PREDICATE
from ._filter import CONTAINS_REGEXP_PREDICATE
from ._filter import ENDSWITH_PREDICATE
from ._filter import EQ_PREDICATE
from ._filter import STARTS_WITH_PREDICATE
from ._filter_predicate import FilterPredicate
from .._file_collection_display import FileCollectionDisplay


class DirectoryPredicate(FilePredicate):
    """A predicate that only filters directories"""

    def __init__(self):
        subpredicates = (
            CONTAINS_PREDICATE,
            CONTAINS_REGEXP_PREDICATE,
            STARTS_WITH_PREDICATE,
            ENDSWITH_PREDICATE,
            EQ_PREDICATE,
        )
        predicates = [
            DoesPredicate(subpredicates),
            DoesNotPredicate(subpredicates),
        ]
        FilterPredicate.__init__(
            self,
            "directory",
            "Directory",
            self.fn_filter,
            predicates,
            doc="Apply the rule to directories",
        )

    @staticmethod
    def fn_filter(node_type__modpath__module, *args):
        """The DirectoryPredicate filter function

        The arg slot expects a tuple of node_type and modpath.
        The predicate returns None (= agnostic about filtering) if
        the node is not a directory, otherwise it composites the
        modpath into a file path and applies it to the rest of
        the args.
        """
        (node_type, modpath, module) = node_type__modpath__module
        if isinstance(modpath[-1], tuple) and len(modpath[-1]) == 3:
            path = os.path.join(*modpath[:-2])
        else:
            path = os.path.join(*modpath[:-1])
        return args[0](path, *args[1:])

    def test_valid(self, pipeline, *args):
        self(
            (FileCollectionDisplay.NODE_FILE, ["/imaging", "image.tif"], None,), *args,
        )



from ._abstract_image import AbstractImage
from ._callback_image import CallbackImage
from ._vanilla_image import VanillaImage
from .file import FileImage
from .file import FlexFrameImage
from .file import MovieFrameImage
from .file import STKFrameImage
from .file import ColorImage
from .file import MaskImage
from .file import MonochromeImage
from .file import ObjectsImage
from .file import URLImage



from .._setting import Setting


class DoSomething(Setting):
    """Do something in response to a button press
    """

    save_to_pipeline = False

    def __init__(self, text, label, callback, *args, **kwargs):
        super(DoSomething, self).__init__(text, "n/a", **kwargs)
        self.__label = label
        self.__callback = callback
        self.__args = args

    def get_label(self):
        """Return the text label for the button"""
        return self.__label

    def set_label(self, label):
        self.__label = label

    label = property(get_label, set_label)

    def on_event_fired(self):
        """Call the callback in response to the user's request to do something"""
        self.__callback(*self.__args)



from ._range import Range


class FloatRange(Range):
    """A setting that allows only floating point input between two constrained values
    """

    valid_format_text = "%s must be a floating-point number"

    def __init__(self, text, value=(0, 1), *args, **kwargs):
        """Initialize an integer range
        text  - helpful text to be displayed to the user
        value - initial default value, a two-tuple as minimum and maximum
        minval - the minimum acceptable value of either
        maxval - the maximum acceptable value of either
        """
        smin, smax = [("%f" % v).rstrip("0") for v in value]
        text_value = ",".join([x + "0" if x.endswith(".") else x for x in (smin, smax)])
        super(FloatRange, self).__init__(text, text_value, *args, **kwargs)

    def str_to_value(self, value_str):
        return float(value_str)

    def value_to_str(self, value):
        return "%f" % value



import re

builtin_readers = {
    "imageio_reader": "ImageIOReader",
    "imageio_reader_v3": "ImageIOReaderV3",
    "ngff_reader": "NGFFReader",
    "bioformats_reader": "BioformatsReader",
    "gcs_reader": "GcsReader",
}
# All successfully loaded reader classes. Maps name:class
ALL_READERS = dict()
# Reader classes that failed to load. Maps name:exception str
BAD_READERS = dict()
# Active reader classes (ALL_READERS that aren't disabled by user). Maps name:class
AVAILABLE_READERS = dict()

ZARR_FILETYPE = re.compile(r"(?<=\.zarr)", flags=re.IGNORECASE)




from ._multichoice import MultiChoice
from ...constants.setting import get_name_provider_choices


class SubscriberMultiChoice(MultiChoice):
    """A multi-choice setting that gets its choices through providers

    This setting operates similarly to the name subscribers. It gets
    its choices from the name providers for the subscriber's group.
    It displays a list of choices and the user can select multiple
    choices.
    """

    def __init__(self, text, group, value=None, *args, **kwargs):
        self.__required_attributes = {"group": group}
        if "required_attributes" in kwargs:
            self.__required_attributes.update(kwargs["required_attributes"])
            kwargs = kwargs.copy()
            del kwargs["required_attributes"]
        super(SubscriberMultiChoice, self).__init__(text, [], value, *args, **kwargs)

    def load_choices(self, pipeline):
        """Get the choice list from name providers"""
        self.choices = sorted(
            [c[0] for c in get_name_provider_choices(pipeline, self, self.group)]
        )

    @property
    def group(self):
        return self.__required_attributes["group"]

    def matches(self, provider):
        """Return true if the provider is compatible with this subscriber

        This method can be used to be more particular about the providers
        that are selected. For instance, if you want a list of only
        FileImageNameProviders (images loaded from files), you can
        check that here.
        """
        return all(
            [
                provider.provided_attributes.get(key, None)
                == self.__required_attributes[key]
                for key in self.__required_attributes
            ]
        )

    def test_valid(self, pipeline):
        self.load_choices(pipeline)
        super(SubscriberMultiChoice, self).test_valid(pipeline)



from ._multichoice import MultiChoice


class MeasurementMultiChoice(MultiChoice):
    """A multi-choice setting for selecting multiple measurements"""

    def __init__(self, text, value="", *args, **kwargs):
        """Initialize the measurement multi-choice

        At initialization, the choices are empty because the measurements
        can't be fetched here. It's done (bit of a hack) in test_valid.
        """
        super(MeasurementMultiChoice, self).__init__(text, [], value, *args, **kwargs)

    @staticmethod
    def encode_object_name(object_name):
        """Encode object name, escaping |"""
        return object_name.replace("|", "||")

    @staticmethod
    def decode_object_name(object_name):
        """Decode the escaped object name"""
        return object_name.replace("||", "|")

    @staticmethod
    def split_choice(choice):
        """Split object and feature within a choice"""
        subst_choice = choice.replace("||", "++")
        loc = subst_choice.find("|")
        if loc == -1:
            return subst_choice, "Invalid"
        return choice[:loc], choice[(loc + 1) :]

    def get_measurement_object(self, choice):
        return self.decode_object_name(self.split_choice(choice)[0])

    def get_measurement_feature(self, choice):
        return self.split_choice(choice)[1]

    def make_measurement_choice(self, object_name, feature):
        return self.encode_object_name(object_name) + "|" + feature

    @staticmethod
    def get_value_string(choices):
        """Return the string value representing the choices made

        choices - a collection of choices as returned by make_measurement_choice
        """
        return ",".join(choices)

    def test_valid(self, pipeline):
        """Get the choices here and call the superclass validator"""
        self.populate_choices(pipeline)
        super(MeasurementMultiChoice, self).test_valid(pipeline)

    def populate_choices(self, pipeline):
        #
        # Find our module
        #
        for module in pipeline.modules():
            for setting in module.visible_settings():
                if id(setting) == id(self):
                    break
        columns = pipeline.get_measurement_columns(module)

        def valid_mc(c):
            """Disallow any measurement column with "," or "|" in its names"""
            return not any([any([bad in f for f in c[:2]]) for bad in (",", "|")])

        self.set_choices(
            [self.make_measurement_choice(c[0], c[1]) for c in columns if valid_mc(c)]
        )



import numpy
import scipy.ndimage

from cellprofiler_core.constants.measurement import (
    M_LOCATION_CENTER_X,
    M_LOCATION_CENTER_Y,
    M_NUMBER_OBJECT_NUMBER,
    FF_COUNT,
    COLTYPE_FLOAT,
    COLTYPE_INTEGER,
    IMAGE,
    M_LOCATION_CENTER_Z,
)


def add_object_location_measurements(
    measurements, object_name, labels, object_count=None
):
    """Add the X and Y centers of mass to the measurements

    measurements - the measurements container
    object_name  - the name of the objects being measured
    labels       - the label matrix
    object_count - (optional) the object count if known, otherwise
                   takes the maximum value in the labels matrix which is
                   usually correct.
    """
    if object_count is None:
        object_count = numpy.max(labels)
    #
    # Get the centers of each object - center_of_mass <- list of two-tuples.
    #
    if object_count:
        centers = scipy.ndimage.center_of_mass(
            numpy.ones(labels.shape), labels, list(range(1, object_count + 1))
        )
        centers = numpy.array(centers)
        centers = centers.reshape((object_count, len(labels.shape)))
        if centers.shape[1] != 3:
            location_center_y = centers[:, 0]
            location_center_x = centers[:, 1]
        else:
            location_center_z = centers[:, 0]
            location_center_y = centers[:, 1]
            location_center_x = centers[:, 2]
        number = numpy.arange(1, object_count + 1)
    else:
        location_center_z = numpy.zeros((0,), dtype=float)
        location_center_y = numpy.zeros((0,), dtype=float)
        location_center_x = numpy.zeros((0,), dtype=float)
        number = numpy.zeros((0,), dtype=int)
    measurements.add_measurement(
        object_name, M_LOCATION_CENTER_X, location_center_x,
    )
    measurements.add_measurement(
        object_name, M_LOCATION_CENTER_Y, location_center_y,
    )
    if len(labels.shape) > 2:
        measurements.add_measurement(
            object_name, M_LOCATION_CENTER_Z, location_center_z,
        )

    measurements.add_measurement(
        object_name, M_NUMBER_OBJECT_NUMBER, number,
    )


def add_object_location_measurements_ijv(
    measurements, object_name, ijv, object_count=None
):
    """Add object location measurements for IJV-style objects"""
    if object_count is None:
        object_count = 0 if ijv.shape[0] == 0 else numpy.max(ijv[:, 2])
    if object_count == 0:
        center_x = numpy.zeros(0)
        center_y = numpy.zeros(0)
    else:
        areas = numpy.zeros(object_count, int)
        areas_bc = numpy.bincount(ijv[:, 2])[1:]
        areas[: len(areas_bc)] = areas_bc
        center_x = numpy.bincount(ijv[:, 2], ijv[:, 1])[1:] / areas
        center_y = numpy.bincount(ijv[:, 2], ijv[:, 0])[1:] / areas
    measurements.add_measurement(
        object_name, M_LOCATION_CENTER_X, center_x,
    )
    measurements.add_measurement(
        object_name, M_LOCATION_CENTER_Y, center_y,
    )
    measurements.add_measurement(
        object_name, M_NUMBER_OBJECT_NUMBER, numpy.arange(1, object_count + 1),
    )


def add_object_count_measurements(measurements, object_name, object_count):
    """Add the # of objects to the measurements"""
    measurements.add_measurement(
        "Image", FF_COUNT % object_name, numpy.array([object_count], dtype=float),
    )


def get_object_measurement_columns(object_name):
    """Get the column definitions for measurements made by identify modules

    Identify modules can use this call when implementing
    CPModule.get_measurement_columns to get the column definitions for
    the measurements made by add_object_location_measurements and
    add_object_count_measurements.
    """
    return [
        (object_name, M_LOCATION_CENTER_X, COLTYPE_FLOAT,),
        (object_name, M_LOCATION_CENTER_Y, COLTYPE_FLOAT,),
        (object_name, M_NUMBER_OBJECT_NUMBER, COLTYPE_INTEGER,),
        (IMAGE, FF_COUNT % object_name, COLTYPE_INTEGER,),
    ]



from ._do_something import DoSomething


class PathListRefreshButton(DoSomething):
    """A setting that displays as a button which refreshes the path list"""

    def __init__(self, text, label, *args, **kwargs):
        DoSomething.__init__(self, text, label, self.fn_callback, *args, **kwargs)
        # callback set by module view
        self.callback = None

    def fn_callback(self, *args, **kwargs):
        if self.callback is not None:
            self.callback(*args, **kwargs)



from ._event import Event


class PrepareRunError(Event):
    """A user configuration error prevented CP from running the pipeline

    Modules use this class to report conditions that prevent construction
    of the image set list. An example would be if the user misconfigured
    LoadImages or NamesAndTypes and no images were matched.
    """

    def __init__(self, module, message):
        super(self.__class__, self).__init__()
        self.module = module
        self.message = message



from ._image_subscriber import ImageSubscriber


class FileImageSubscriber(ImageSubscriber):
    """A setting that provides image names loaded from files"""

    def __init__(
        self,
        text,
        value="Do not use",
        can_be_blank=False,
        blank_text="Leave blank",
        *args,
        **kwargs,
    ):
        kwargs = kwargs.copy()
        if "required_attributes" not in kwargs:
            kwargs["required_attributes"] = {}
        kwargs["required_attributes"]["file_image"] = True
        super(FileImageSubscriber, self).__init__(
            text, value, can_be_blank, blank_text, *args, **kwargs
        )



import functools

from .._setting import Setting
from .._validation_error import ValidationError


class MultiChoice(Setting):
    """A setting that represents selection of multiple choices from a list"""

    def __init__(self, text, choices, value=None, *args, **kwargs):
        """Initializer

        text - the explanatory text for the setting
        choices - a sequence of string choices to be selected
        value - a list of selected choices or a comma-separated string list
        """
        super(MultiChoice, self).__init__(
            text, self.parse_value(value), *args, **kwargs
        )
        self.__choices = choices

    @staticmethod
    def parse_value(value):
        if value is None:
            return ""
        elif isinstance(value, str):
            return value
        elif hasattr(value, "__getitem__"):
            return ",".join(value)
        raise ValueError("Unexpected value type: %s" % type(value))

    def __internal_get_choices(self):
        """The sequence of strings that define the choices to be displayed"""
        return self.get_choices()

    def get_choices(self):
        """The sequence of strings that define the choices to be displayed"""
        return self.__choices

    def __internal_set_choices(self, choices):
        return self.set_choices(choices)

    def set_choices(self, choices):
        self.__choices = choices

    choices = property(__internal_get_choices, __internal_set_choices)

    def set_value(self, value):
        """Set the value of a multi-choice setting

        value is either a single string, a comma-separated string of
        multiple choices or a list of strings
        """
        super(MultiChoice, self).set_value(self.parse_value(value))

    def get_selections(self):
        """Return the currently selected values"""
        value = self.get_value()
        if len(value) == 0:
            return ()
        return value.split(",")

    selections = property(get_selections)

    def test_valid(self, pipeline):
        """Ensure that the selections are among the choices"""
        for selection in self.get_selections():
            if selection not in self.choices:
                if len(self.choices) == 0:
                    raise ValidationError("No available choices", self)
                elif len(self.choices) > 25:
                    raise ValidationError(
                        "%s is not one of the choices" % selection, self
                    )
                raise ValidationError(
                    "%s is not one of %s"
                    % (
                        selection,
                        functools.reduce(lambda x, y: "%s,%s" % (x, y), self.choices),
                    ),
                    self,
                )



from cellprofiler_core.utilities.zmq.communicable.request._request import Request


class LockStatusRequest(Request):
    """A request for the status on some locked file

    uid - the unique ID stored inside the file's lock
    """

    def __init__(self, uid, **kwargs):
        self.uid = uid
        Request.__init__(self, **kwargs)



# coding=utf-8

"""
Align
=====

**Align** aligns images relative to each other, for example, to correct
shifts in the optical path of a microscope in each channel of a
multi-channel set of images.

For two or more input images, this module determines the optimal
alignment among them. Aligning images is useful to obtain proper
measurements of the intensities in one channel based on objects
identified in another channel, for example. Alignment is often needed
when the microscope is not perfectly calibrated. It can also be useful
to align images in a time-lapse series of images. The module stores the
amount of shift between images as a measurement, which can be useful for
quality control purposes.

Note that the second image (and others following) is always aligned with
respect to the first image. That is, the X/Y offsets indicate how much
the second image needs to be shifted by to match the first.

This module does not perform warping or rotation, it simply shifts images
in X and Y. For more complex registration tasks, you might preprocess
images using a plugin for that purpose in FIJI/ImageJ.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          NO           YES
============ ============ ===============



Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Xshift, Yshift:* The pixel shift in X and Y of the aligned image
   with respect to the original image.

References
^^^^^^^^^^

-  Lewis JP. (1995) “Fast normalized cross-correlation.” *Vision
   Interface*, 1-7.
"""

import numpy as np
import scipy.ndimage as scind
import scipy.sparse
from centrosome.filter import stretch
from scipy.fftpack import fft2, ifft2

from ..constants.measurement import COLTYPE_INTEGER
from ..image import Image
from ..module import Module
from ..setting import Divider, SettingsGroup
from ..setting.choice import Choice
from ..setting.do_something import DoSomething, RemoveSettingButton
from ..setting.subscriber import ImageSubscriber
from ..setting.text import ImageName

M_MUTUAL_INFORMATION = "Mutual Information"
M_CROSS_CORRELATION = "Normalized Cross Correlation"
M_ALL = (M_MUTUAL_INFORMATION, M_CROSS_CORRELATION)

A_SIMILARLY = "Similarly"
A_SEPARATELY = "Separately"

C_SAME_SIZE = "Keep size"
C_CROP = "Crop to aligned region"
C_PAD = "Pad images"

C_ALIGN = "Align"

MEASUREMENT_FORMAT = C_ALIGN + "_%sshift_%s"


class Align(Module):
    module_name = "Align"
    category = "Image Processing"
    variable_revision_number = 3

    def create_settings(self):
        self.first_input_image = ImageSubscriber(
            "Select the first input image",
            "None",
            doc="""\
Specify the name of the first image to align.""",
        )

        self.first_output_image = ImageName(
            "Name the first output image",
            "AlignedRed",
            doc="""\
Enter the name of the first aligned image.""",
        )

        self.separator_1 = Divider(line=False)
        self.second_input_image = ImageSubscriber(
            "Select the second input image",
            "None",
            doc="""\
Specify the name of the second image to align.""",
        )

        self.second_output_image = ImageName(
            "Name the second output image",
            "AlignedGreen",
            doc="""\
Enter the name of the second aligned image.""",
        )

        self.separator_2 = Divider(line=False)
        self.additional_images = []
        self.add_button = DoSomething("", "Add another image", self.add_image)

        self.alignment_method = Choice(
            "Select the alignment method",
            M_ALL,
            doc="""\
Two options for the alignment method are available:

-  *%(M_MUTUAL_INFORMATION)s:* This more general method works well for
   aligning images from different modalities that contain the same
   information, but are expressed differently. However, this method
   performs better than %(M_CROSS_CORRELATION)s, even in the same
   modality, if the images are not highly correlated. It is iterative,
   and thus tends to be slower than other methods, but is more likely to
   be correct. Essentially, alignment is performed by measuring how well
   one image “explains” the other. For example, a fluorescent image can
   be aligned to a brightfield image by this method since the relevant
   features are bright in one modality where they are dim in the other.
-  *%(M_CROSS_CORRELATION)s:* This is a good means of alignment in the
   case of images acquired with the same modality (e.g., all images to
   be aligned are fluorescent). It is fast, however it can be highly
   influenced by a particular, possibly spurious, feature and in turn
   generate anomalously large shifts. It allows for a linear
   relationship between the intensities of the two images, i.e., the
   relevant features in the images to be aligned all have varying
   degrees of brightness.

"""
                % globals(),
        )

        self.crop_mode = Choice(
            "Crop mode",
            [C_CROP, C_PAD, C_SAME_SIZE],
            doc="""\
The crop mode determines how the output images are either cropped or
padded after alignment. The alignment phase calculates the areas in each
image that are found to be overlapping. In almost all cases, there will
be portions of some or all of the images that don’t overlap with any
other aligned image. These portions have no counterpart and will be
excluded from analysis. There are three choices for cropping:

-  *%(C_CROP)s:* Crop every image to the region that overlaps in all
   images. This makes downstream analysis simpler because all of the
   output images have authentic pixel data at all positions, however it
   discards parts of images. Also, the output images may not be the same
   size as the input images which may cause problems if downstream
   modules use aligned and unaligned images (which may be of differing
   sizes) in combination.
-  *%(C_PAD)s:* Align every image and pad with masked black pixels to
   make each image the same size. This results in larger images, but
   preserves all information in each of the images. This may be the best
   choice if images undergo an operation such as smoothing that could
   use the information that would otherwise be cropped.
-  *%(C_SAME_SIZE)s:* Maintain the sizes of the images but align them,
   masking the unaligned portions with black pixels. **Align** aligns
   all images relative to the first. This is a reasonable option for
   alignments with small displacements since it maintains a consistent
   image size which may be useful if output images from different image
   sets will be compared against each other after processing. The
   reference image can also be used across image sets. For example, the
   reference image could be loaded for all image sets in a group to
   align the entire group’s images similarly, then the aligned images
   could be combined in a module such as **MakeProjection**.
   """
                % globals(),
        )



    def add_image(self, can_remove=True):
        """Add an image + associated questions and buttons"""
        group = SettingsGroup()
        if can_remove:
            group.append("divider", Divider(line=False))

        group.append(
            "input_image_name",
            ImageSubscriber(
                "Select the additional image",
                "None",
                doc="""
 Select the additional image to align?""",
            ),
        )

        group.append(
            "output_image_name",
            ImageName(
                "Name the output image",
                "AlignedBlue",
                doc="""
 Enter the name of the aligned image?""",
            ),
        )

        group.append(
            "align_choice",
            Choice(
                "Select how the alignment is to be applied",
                [A_SIMILARLY, A_SEPARATELY],
                doc="""\
An additional image can either be aligned similarly to the second one or
a separate alignment to the first image can be calculated:

-  *%(A_SIMILARLY)s:* The same alignment measurements obtained from the
   first two input images are applied to this additional image.
-  *%(A_SEPARATELY)s:* A new set of alignment measurements are
   calculated for this additional image using the alignment method
   specified with respect to the first input image.
"""
                    % globals(),
            ),
        )

        if can_remove:
            group.append(
                "remover",
                RemoveSettingButton(
                    "", "Remove above image", self.additional_images, group
                ),
            )
        self.additional_images.append(group)

    def settings(self):

        result = [self.alignment_method, self.crop_mode]

        result += [
            self.first_input_image,
            self.first_output_image,
            self.second_input_image,
            self.second_output_image,
        ]
        for additional in self.additional_images:
            result += [
                additional.input_image_name,
                additional.output_image_name,
                additional.align_choice,
            ]
        return result

    def prepare_settings(self, setting_values):
        assert (len(setting_values) - 6) % 3 == 0
        n_additional = (len(setting_values) - 6) / 3
        del self.additional_images[:]
        while len(self.additional_images) < n_additional:
            self.add_image()

    def visible_settings(self):
        result = [self.alignment_method, self.crop_mode]

        result += [
            self.first_input_image,
            self.first_output_image,
            self.separator_1,
            self.second_input_image,
            self.second_output_image,
        ]
        for additional in self.additional_images:
            result += additional.visible_settings()
        result += [self.add_button]
        return result

    def run(self, workspace):
        i_min = np.iinfo(int).max
        j_min = np.iinfo(int).max
        off_x, off_y = self.align(
            workspace, self.first_input_image.value, self.second_input_image.value
        )
        names = [
            (self.first_input_image.value, self.first_output_image.value),
            (self.second_input_image.value, self.second_output_image.value),
        ]
        offsets = [(0, 0), (off_y, off_x)]

        for additional in self.additional_images:
            names.append(
                (additional.input_image_name.value, additional.output_image_name.value)
            )
            if additional.align_choice == A_SIMILARLY:
                a_off_x, a_off_y = off_x, off_y
            else:
                a_off_x, a_off_y = self.align(
                    workspace,
                    self.first_input_image.value,
                    additional.input_image_name.value,
                )
            offsets.append((a_off_y, a_off_x))

        shapes = [
            workspace.image_set.get_image(x).pixel_data.shape[:2] for x, _ in names
        ]
        offsets, shapes = self.adjust_offsets(offsets, shapes)

        #
        # Align and write the measurements
        #
        for (input_name, output_name), (y, x), shape in zip(names, offsets, shapes):
            self.apply_alignment(workspace, input_name, output_name, x, y, shape)
            for axis, value in (("X", -x), ("Y", -y)):
                feature = MEASUREMENT_FORMAT % (axis, output_name)
                workspace.measurements.add_image_measurement(feature, value)

        # save data for display
        workspace.display_data.image_info = [
            (
                input_name,
                workspace.image_set.get_image(input_name).pixel_data,
                output_name,
                workspace.image_set.get_image(output_name).pixel_data,
                x,
                y,
                shape,
            )
            for (input_name, output_name), (y, x), shape in zip(names, offsets, shapes)
        ]

    def display(self, workspace, figure):
        """Display the overlaid images

        workspace - the workspace being run, with display_data holding:
            image_info - a list of lists:
                 input image name of image being aligned
                 input image data
                 output image name of image being aligned
                 output image data
                 x offset
                 y offset
        """
        image_info = workspace.display_data.image_info
        first_input_name = self.first_input_image.value
        first_output_name = self.first_output_image.value
        figure.set_subplots((2, len(image_info) - 1))

        first_input_pixels = image_info[0][1]
        first_output_pixels = image_info[0][3]
        for (
                j,
                (input_name, input_pixels, output_name, output_pixels, off_x, off_y, shape),
        ) in enumerate(image_info[1:]):
            unaligned_title = "Unaligned images: %s and %s" % (
                first_input_name,
                input_name,
            )
            #
            # Make them grayscale if needed
            #
            first_pixels, other_pixels = [
                img if img.ndim == 2 else np.mean(img, 2)
                for img in (first_input_pixels, input_pixels)
            ]
            max_shape = np.maximum(first_pixels.shape, other_pixels.shape)
            img = np.zeros((max_shape[0], max_shape[1], 3))
            img[: first_pixels.shape[0], : first_pixels.shape[1], 0] = first_pixels
            img[: other_pixels.shape[0], : other_pixels.shape[1], 1] = other_pixels
            figure.subplot_imshow(
                0, j, img, unaligned_title, sharexy=figure.subplot(0, 0)
            )

            aligned_title = "Aligned images: %s and %s\nX offset: %d, Y offset: %d" % (
                first_output_name,
                output_name,
                -off_x,
                -off_y,
            )
            first_pixels, other_pixels = [
                img if img.ndim == 2 else np.mean(img, 2)
                for img in (first_output_pixels, output_pixels)
            ]
            max_shape = np.maximum(first_pixels.shape, other_pixels.shape)
            img = np.zeros((max_shape[0], max_shape[1], 3))
            img[: first_pixels.shape[0], : first_pixels.shape[1], 0] = first_pixels
            img[: other_pixels.shape[0], : other_pixels.shape[1], 1] = other_pixels
            figure.subplot_imshow(
                1, j, img, aligned_title, sharexy=figure.subplot(0, 0)
            )

    def align(self, workspace, input1_name, input2_name):
        """Align the second image with the first

        Calculate the alignment offset that must be added to indexes in the
        first image to arrive at indexes in the second image.

        Returns the x,y (not i,j) offsets.
        """
        image1 = workspace.image_set.get_image(input1_name)
        image1_pixels = image1.pixel_data.astype(float)
        image2 = workspace.image_set.get_image(input2_name)
        image2_pixels = image2.pixel_data.astype(float)
        if self.alignment_method == M_CROSS_CORRELATION:
            return self.align_cross_correlation(image1_pixels, image2_pixels)
        else:
            image1_mask = image1.mask
            image2_mask = image2.mask
            return self.align_mutual_information(
                image1_pixels, image2_pixels, image1_mask, image2_mask
            )

    @staticmethod
    def align_cross_correlation(pixels1, pixels2):
        """Align the second image with the first using max cross-correlation

        returns the x,y offsets to add to image1's indexes to align it with
        image2

        Many of the ideas here are based on the paper, "Fast Normalized
        Cross-Correlation" by J.P. Lewis
        (http://www.idiom.com/~zilla/Papers/nvisionInterface/nip.html)
        which is frequently cited when addressing this problem.
        """
        #
        # TODO: Possibly use all 3 dimensions for color some day
        #
        if pixels1.ndim == 3:
            pixels1 = np.mean(pixels1, 2)
        if pixels2.ndim == 3:
            pixels2 = np.mean(pixels2, 2)
        #
        # We double the size of the image to get a field of zeros
        # for the parts of one image that don't overlap the displaced
        # second image.
        #
        # Since we're going into the frequency domain, if the images are of
        # different sizes, we can make the FFT shape large enough to capture
        # the period of the largest image - the smaller just will have zero
        # amplitude at that frequency.
        #
        s = np.maximum(pixels1.shape, pixels2.shape)
        fshape = s * 2
        #
        # Calculate the # of pixels at a particular point
        #
        i, j = np.mgrid[-s[0]: s[0], -s[1]: s[1]]
        unit = np.abs(i * j).astype(float)
        unit[unit < 1] = 1  # keeps from dividing by zero in some places
        #
        # Normalize the pixel values around zero which does not affect the
        # correlation, keeps some of the sums of multiplications from
        # losing precision and precomputes t(x-u,y-v) - t_mean
        #
        pixels1 = pixels1 - np.mean(pixels1)
        pixels2 = pixels2 - np.mean(pixels2)
        #
        # Lewis uses an image, f and a template t. He derives a normalized
        # cross correlation, ncc(u,v) =
        # sum((f(x,y)-f_mean(u,v))*(t(x-u,y-v)-t_mean),x,y) /
        # sqrt(sum((f(x,y)-f_mean(u,v))**2,x,y) * (sum((t(x-u,y-v)-t_mean)**2,x,y)
        #
        # From here, he finds that the numerator term, f_mean(u,v)*(t...) is zero
        # leaving f(x,y)*(t(x-u,y-v)-t_mean) which is a convolution of f
        # by t-t_mean.
        #
        fp1 = fft2(pixels1, fshape.tolist())
        fp2 = fft2(pixels2, fshape.tolist())
        corr12 = ifft2(fp1 * fp2.conj()).real

        #
        # Use the trick of Lewis here - compute the cumulative sums
        # in a fashion that accounts for the parts that are off the
        # edge of the template.
        #
        # We do this in quadrants:
        # q0 q1
        # q2 q3
        # For the first,
        # q0 is the sum over pixels1[i:,j:] - sum i,j backwards
        # q1 is the sum over pixels1[i:,:j] - sum i backwards, j forwards
        # q2 is the sum over pixels1[:i,j:] - sum i forwards, j backwards
        # q3 is the sum over pixels1[:i,:j] - sum i,j forwards
        #
        # The second is done as above but reflected lr and ud
        #
        p1_si = pixels1.shape[0]
        p1_sj = pixels1.shape[1]
        p1_sum = np.zeros(fshape)
        p1_sum[:p1_si, :p1_sj] = cumsum_quadrant(pixels1, False, False)
        p1_sum[:p1_si, -p1_sj:] = cumsum_quadrant(pixels1, False, True)
        p1_sum[-p1_si:, :p1_sj] = cumsum_quadrant(pixels1, True, False)
        p1_sum[-p1_si:, -p1_sj:] = cumsum_quadrant(pixels1, True, True)
        #
        # Divide the sum over the # of elements summed-over
        #
        p1_mean = p1_sum / unit

        p2_si = pixels2.shape[0]
        p2_sj = pixels2.shape[1]
        p2_sum = np.zeros(fshape)
        p2_sum[:p2_si, :p2_sj] = cumsum_quadrant(pixels2, False, False)
        p2_sum[:p2_si, -p2_sj:] = cumsum_quadrant(pixels2, False, True)
        p2_sum[-p2_si:, :p2_sj] = cumsum_quadrant(pixels2, True, False)
        p2_sum[-p2_si:, -p2_sj:] = cumsum_quadrant(pixels2, True, True)
        p2_sum = np.fliplr(np.flipud(p2_sum))
        p2_mean = p2_sum / unit
        #
        # Once we have the means for u,v, we can calculate the
        # variance-like parts of the equation. We have to multiply
        # the mean^2 by the # of elements being summed-over
        # to account for the mean being summed that many times.
        #
        p1sd = np.sum(pixels1 ** 2) - p1_mean ** 2 * np.product(s)
        p2sd = np.sum(pixels2 ** 2) - p2_mean ** 2 * np.product(s)
        #
        # There's always chance of roundoff error for a zero value
        # resulting in a negative sd, so limit the sds here
        #
        sd = np.sqrt(np.maximum(p1sd * p2sd, 0))
        corrnorm = corr12 / sd
        #
        # There's not much information for points where the standard
        # deviation is less than 1/100 of the maximum. We exclude these
        # from consideration.
        #
        corrnorm[(unit < np.product(s) / 2) & (sd < np.mean(sd) / 100)] = 0
        i, j = np.unravel_index(np.argmax(corrnorm), fshape)
        #
        # Reflect values that fall into the second half
        #
        if i > pixels1.shape[0]:
            i = i - fshape[0]
        if j > pixels1.shape[1]:
            j = j - fshape[1]
        return j, i

    @staticmethod
    def align_mutual_information(pixels1, pixels2, mask1, mask2):
        """Align the second image with the first using mutual information

        returns the x,y offsets to add to image1's indexes to align it with
        image2

        The algorithm computes the mutual information content of the two
        images, offset by one in each direction (including diagonal) and
        then picks the direction in which there is the most mutual information.
        From there, it tries all offsets again and so on until it reaches
        a local maximum.
        """
        #
        # TODO: Possibly use all 3 dimensions for color some day
        #
        if pixels1.ndim == 3:
            pixels1 = np.mean(pixels1, 2)
        if pixels2.ndim == 3:
            pixels2 = np.mean(pixels2, 2)

        def mutualinf(x, y, maskx, masky):
            x = x[maskx & masky]
            y = y[maskx & masky]
            return entropy(x) + entropy(y) - entropy2(x, y)

        maxshape = np.maximum(pixels1.shape, pixels2.shape)
        pixels1 = reshape_image(pixels1, maxshape)
        pixels2 = reshape_image(pixels2, maxshape)
        mask1 = reshape_image(mask1, maxshape)
        mask2 = reshape_image(mask2, maxshape)

        best = mutualinf(pixels1, pixels2, mask1, mask2)
        i = 0
        j = 0
        while True:
            last_i = i
            last_j = j
            for new_i in range(last_i - 1, last_i + 2):
                for new_j in range(last_j - 1, last_j + 2):
                    if new_i == 0 and new_j == 0:
                        continue
                    p2, p1 = offset_slice(pixels2, pixels1, new_i, new_j)
                    m2, m1 = offset_slice(mask2, mask1, new_i, new_j)
                    info = mutualinf(p1, p2, m1, m2)
                    if info > best:
                        best = info
                        i = new_i
                        j = new_j
            if i == last_i and j == last_j:
                return j, i

    @staticmethod
    def apply_alignment(
            workspace, input_image_name, output_image_name, off_x, off_y, shape
    ):
        """Apply an alignment to the input image to result in the output image

        workspace - image set's workspace passed to run

        input_image_name - name of the image to be aligned

        output_image_name - name of the resultant image

        off_x, off_y - offset of the resultant image relative to the original

        shape - shape of the resultant image
        """

        image = workspace.image_set.get_image(input_image_name)
        pixel_data = image.pixel_data
        if pixel_data.ndim == 2:
            output_shape = (shape[0], shape[1], 1)
            planes = [pixel_data]
        else:
            output_shape = (shape[0], shape[1], pixel_data.shape[2])
            planes = [pixel_data[:, :, i] for i in range(pixel_data.shape[2])]
        output_pixels = np.zeros(output_shape, pixel_data.dtype)
        for i, plane in enumerate(planes):
            #
            # Copy the input to the output
            #
            p1, p2 = offset_slice(plane, output_pixels[:, :, i], off_y, off_x)
            p2[:, :] = p1[:, :]
        if pixel_data.ndim == 2:
            output_pixels.shape = output_pixels.shape[:2]
        output_mask = np.zeros(shape, bool)
        p1, p2 = offset_slice(image.mask, output_mask, off_y, off_x)
        p2[:, :] = p1[:, :]
        if np.all(output_mask):
            output_mask = None
        crop_mask = np.zeros(image.pixel_data.shape, bool)
        p1, p2 = offset_slice(crop_mask, output_pixels, off_y, off_x)
        p1[:, :] = True
        if np.all(crop_mask):
            crop_mask = None
        output_image = Image(
            output_pixels, mask=output_mask, crop_mask=crop_mask, parent_image=image
        )
        workspace.image_set.add(output_image_name, output_image)

    def adjust_offsets(self, offsets, shapes):
        """Adjust the offsets and shapes for output

        workspace - workspace passed to "run"

        offsets - i,j offsets for each image

        shapes - shapes of the input images

        names - pairs of input / output names

        Based on the crop mode, adjust the offsets and shapes to optimize
        the cropping.
        """
        offsets = np.array(offsets)
        shapes = np.array(shapes)
        if self.crop_mode == C_CROP:
            # modify the offsets so that all are negative
            max_offset = np.max(offsets, 0)
            offsets = offsets - max_offset[np.newaxis, :]
            #
            # Reduce each shape by the amount chopped off
            #
            shapes += offsets
            #
            # Pick the smallest in each of the dimensions and repeat for all
            #
            shape = np.min(shapes, 0)
            shapes = np.tile(shape, len(shapes))
            shapes.shape = offsets.shape
        elif self.crop_mode == C_PAD:
            #
            # modify the offsets so that they are all positive
            #
            min_offset = np.min(offsets, 0)
            offsets = offsets - min_offset[np.newaxis, :]
            #
            # Expand each shape by the top-left padding
            #
            shapes += offsets
            #
            # Pick the largest in each of the dimensions and repeat for all
            #
            shape = np.max(shapes, 0)
            shapes = np.tile(shape, len(shapes))
            shapes.shape = offsets.shape
        return offsets.tolist(), shapes.tolist()

    def get_categories(self, pipeline, object_name):
        if object_name == "Image":
            return [C_ALIGN]
        return []

    def get_measurements(self, pipeline, object_name, category):
        if object_name == "Image" and category == C_ALIGN:
            return ["Xshift", "Yshift"]
        return []

    def get_measurement_images(self, pipeline, object_name, category, measurement):
        if measurement in self.get_measurements(pipeline, object_name, category):
            return [self.first_output_image.value, self.second_output_image.value] + [
                additional.output_image_name.value
                for additional in self.additional_images
            ]
        return []

    def get_measurement_columns(self, pipeline):
        """return the offset measurements"""

        targets = [self.first_output_image.value, self.second_output_image.value] + [
            additional.output_image_name.value for additional in self.additional_images
        ]
        columns = []
        for axis in ("X", "Y"):
            columns += [
                ("Image", MEASUREMENT_FORMAT % (axis, target), COLTYPE_INTEGER,)
                for target in targets
            ]
        return columns

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):
        if variable_revision_number == 1:
            # Moved final settings (alignment method, cropping) to the top
            setting_values = setting_values[-2:] + setting_values[:-2]
            variable_revision_number = 2

        if variable_revision_number == 2:
            # wants_cropping changed to crop_mode
            setting_values = (
                    setting_values[:1]
                    + [C_CROP if setting_values[1] == "Yes" else C_SAME_SIZE]
                    + setting_values[2:]
            )
            variable_revision_number = 3

        return setting_values, variable_revision_number


def offset_slice(pixels1, pixels2, i, j):
    """Return two sliced arrays where the first slice is offset by i,j
    relative to the second slice.

    """
    if i < 0:
        height = min(pixels1.shape[0] + i, pixels2.shape[0])
        p1_imin = -i
        p2_imin = 0
    else:
        height = min(pixels1.shape[0], pixels2.shape[0] - i)
        p1_imin = 0
        p2_imin = i
    p1_imax = p1_imin + height
    p2_imax = p2_imin + height
    if j < 0:
        width = min(pixels1.shape[1] + j, pixels2.shape[1])
        p1_jmin = -j
        p2_jmin = 0
    else:
        width = min(pixels1.shape[1], pixels2.shape[1] - j)
        p1_jmin = 0
        p2_jmin = j
    p1_jmax = p1_jmin + width
    p2_jmax = p2_jmin + width

    p1 = pixels1[p1_imin:p1_imax, p1_jmin:p1_jmax]
    p2 = pixels2[p2_imin:p2_imax, p2_jmin:p2_jmax]
    return p1, p2


def cumsum_quadrant(x, i_forwards, j_forwards):
    """Return the cumulative sum going in the i, then j direction

    x - the matrix to be summed
    i_forwards - sum from 0 to end in the i direction if true
    j_forwards - sum from 0 to end in the j direction if true
    """
    if i_forwards:
        x = x.cumsum(0)
    else:
        x = np.flipud(np.flipud(x).cumsum(0))
    if j_forwards:
        return x.cumsum(1)
    else:
        return np.fliplr(np.fliplr(x).cumsum(1))


def entropy(x):
    """The entropy of x as if x is a probability distribution"""
    histogram = scind.histogram(x.astype(float), np.min(x), np.max(x), 256)
    n = np.sum(histogram)
    if n > 0 and np.max(histogram) > 0:
        histogram = histogram[histogram != 0]
        return np.log2(n) - np.sum(histogram * np.log2(histogram)) / n
    else:
        return 0


def entropy2(x, y):
    """Joint entropy of paired samples X and Y"""
    #
    # Bin each image into 256 gray levels
    #
    x = (stretch(x) * 255).astype(int)
    y = (stretch(y) * 255).astype(int)
    #
    # create an image where each pixel with the same X & Y gets
    # the same value
    #
    xy = 256 * x + y
    xy = xy.flatten()
    sparse = scipy.sparse.coo_matrix(
        (np.ones(xy.shape, dtype=np.int32), (xy, np.zeros(xy.shape, dtype=np.int32)))
    )
    histogram = sparse.toarray()
    n = np.sum(histogram)
    if n > 0 and np.max(histogram) > 0:
        histogram = histogram[histogram > 0]
        return np.log2(n) - np.sum(histogram * np.log2(histogram)) / n
    else:
        return 0


def reshape_image(source, new_shape):
    """Reshape an image to a larger shape, padding with zeros"""
    if tuple(source.shape) == tuple(new_shape):
        return source

    result = np.zeros(new_shape, source.dtype)
    result[: source.shape[0], : source.shape[1]] = source
    return result



from ._grid_subscriber import GridSubscriber
from ._label_subscriber import LabelSubscriber
from ._subscriber import Subscriber
from .image_subscriber import CropImageSubscriber
from .image_subscriber import FileImageSubscriber
from .image_subscriber import ImageSubscriber
from .image_subscriber import OutlineImageSubscriber
from .list_subscriber import ImageListSubscriber
from .list_subscriber import LabelListSubscriber
from .list_subscriber import ListSubscriber



import io
import pickle

from ._image_set import ImageSet
from ..utilities.image import make_dictionary_key


class ImageSetList:
    """Represents the list of image sets in a pipeline run

    """

    def __init__(self, test_mode=False):
        self.__image_sets = []
        self.__image_sets_by_key = {}
        self.legacy_fields = {}
        self.__associating_by_key = None
        self.combine_path_and_file = False
        self.test_mode = test_mode

    def add_provider_to_all_image_sets(self, provider):
        """Provide an image to every image set

        provider - an instance of AbstractImageProvider
        """
        for image_set in self.__image_sets:
            image_set.add_provider(provider)

    def count(self):
        return len(self.__image_sets)

    def get_groupings(self, keys):
        """Return the groupings of an image set list over a set of keys

        keys - a sequence of keys that match some of the image set keys

        returns an object suitable for use by CPModule.get_groupings:
        tuple of keys, groupings
        keys - the keys as passed into the function
        groupings - a sequence of groupings of image sets where
                    each element of the sequence is a two-tuple.
                    The first element of the two-tuple is a dictionary
                    that gives the group's values for each key.
                    The second element is a list of image numbers of
                    the images in the group
        """
        #
        # Sort order for dictionary keys
        #
        sort_order = []
        dictionaries = []
        #
        # Dictionary of key_values to list of image numbers
        #
        d = {}
        for i in range(self.count()):
            image_set = self.get_image_set(i)
            assert isinstance(image_set, ImageSet)
            key_values = tuple([str(image_set.keys[key]) for key in keys])
            if key_values not in d:
                d[key_values] = []
                sort_order.append(key_values)
            d[key_values].append(i + 1)
        return keys, [(dict(list(zip(keys, k))), d[k]) for k in sort_order]

    def get_image_set(self, keys_or_number):
        """Return either the indexed image set (keys_or_number = index) or the image set with matching keys

        """
        if not isinstance(keys_or_number, dict):
            keys = {"number": keys_or_number}
            number = keys_or_number
            if self.__associating_by_key is None:
                self.__associating_by_key = False
            k = make_dictionary_key(keys)
        else:
            keys = keys_or_number
            k = make_dictionary_key(keys)
            if k in self.__image_sets_by_key:
                number = self.__image_sets_by_key[k].number
            else:
                number = len(self.__image_sets)
            self.__associating_by_key = True
        if number >= len(self.__image_sets):
            self.__image_sets += [None] * (number - len(self.__image_sets) + 1)
        if self.__image_sets[number] is None:
            image_set = ImageSet(number, keys, self.legacy_fields)
            self.__image_sets[number] = image_set
            self.__image_sets_by_key[k] = image_set
            if self.__associating_by_key:
                k = make_dictionary_key(dict(number=number))
                self.__image_sets_by_key[k] = image_set
        else:
            image_set = self.__image_sets[number]
        return image_set

    def load_state(self, state):
        """Load an image_set_list's state from the string returned from save_state"""

        self.__image_sets = []
        self.__image_sets_by_key = {}

        # Make a safe unpickler
        p = pickle.Unpickler(io.BytesIO(state))

        count = p.load()

        all_keys = [p.load() for i in range(count)]

        self.legacy_fields = p.load()

        #
        # Have to do in this order in order for the image set's
        # legacy_fields property to hook to the right legacy_fields
        #
        for i in range(count):
            self.get_image_set(all_keys[i])

    def purge_image_set(self, number):
        """Remove the memory associated with an image set"""
        keys = self.__image_sets[number].keys
        image_set = self.__image_sets[number]
        for provider in image_set.providers:
            provider.release_memory()
        self.__image_sets[number] = None
        self.__image_sets_by_key[repr(keys)] = None

    def save_state(self):
        """Return a string that can be used to load the image_set_list's state

        load_state will restore the image set list's state. No image_set can
        have image providers before this call.
        """
        f = io.BytesIO()
        pickle.dump(self.count(), f)
        for i in range(self.count()):
            image_set = self.get_image_set(i)
            assert isinstance(image_set, ImageSet)
            assert (
                len(image_set.providers) == 0
            ), "An image set cannot have providers while saving its state"
            pickle.dump(image_set.keys, f)
        pickle.dump(self.legacy_fields, f)
        return f.getvalue()



from cellprofiler_core.utilities.zmq.communicable._communicable import Communicable


class Reply(Communicable):
    """The counterpart to a Request.

    All subclasses must accept keyword arguments to __init__() corresponding to
    their attributes.
    """

    def __init__(self, **kwargs):
        # all keywords become attributes
        self.__dict__.update(kwargs)



O_TWO_CLASS = "Two classes"
O_THREE_CLASS = "Three classes"
O_WEIGHTED_VARIANCE = "Weighted variance"
O_ENTROPY = "Entropy"
O_FOREGROUND = "Foreground"
O_BACKGROUND = "Background"
FI_IMAGE_SIZE = "Image size"
FI_CUSTOM = "Custom"
RB_DEFAULT = "Default"
RB_CUSTOM = "Custom"
RB_MEAN = "Mean"
RB_MEDIAN = "Median"
RB_MODE = "Mode"
RB_SD = "Standard deviation"
RB_MAD = "Median absolute deviation"
TS_AUTOMATIC = "Automatic"
TS_GLOBAL = "Global"
TS_ADAPTIVE = "Adaptive"
TS_PER_OBJECT = "Per object"
TS_MANUAL = "Manual"
TS_BINARY_IMAGE = "Binary image"
TS_MEASUREMENT = "Measurement"
TS_ALL = [TS_GLOBAL, TS_ADAPTIVE, TS_MANUAL, TS_MEASUREMENT]
O_FROM_IMAGE = "From image"
TSM_NONE = "No smoothing"
TSM_AUTOMATIC = "Automatic"
TSM_MANUAL = "Manual"
TECH_NOTE_ICON = "gear.png"



AGG_MEAN = "Mean"
AGG_STD_DEV = "StDev"
AGG_MEDIAN = "Median"
AGG_NAMES = [AGG_MEAN, AGG_MEDIAN, AGG_STD_DEV]
IMAGE = "Image"
EXPERIMENT = "Experiment"
RELATIONSHIP = "Relationship"
DB_TEMP = "ExportToDb"
NEIGHBORS = "Neighbors"
OBJECT = "Object"
disallowed_object_names = [IMAGE, EXPERIMENT, RELATIONSHIP]
COLTYPE_INTEGER = "integer"
COLTYPE_FLOAT = "float"
COLTYPE_BLOB = "blob"
COLTYPE_MEDIUMBLOB = "mediumblob"
COLTYPE_LONGBLOB = "longblob"
COLTYPE_VARCHAR_FORMAT = "varchar(%d)"
COLTYPE_VARCHAR = "varchar"
PATH_NAME_LENGTH = 256
FILE_NAME_LENGTH = 128
COLTYPE_VARCHAR_FILE_NAME = COLTYPE_VARCHAR_FORMAT % FILE_NAME_LENGTH
COLTYPE_VARCHAR_PATH_NAME = COLTYPE_VARCHAR_FORMAT % PATH_NAME_LENGTH
MCA_AVAILABLE_EACH_CYCLE = "AvailableEachCycle"
MCA_AVAILABLE_POST_GROUP = "AvailablePostGroup"
MCA_AVAILABLE_POST_RUN = "AvailablePostRun"
C_METADATA = "Metadata"
FTR_SITE = "Site"
FTR_WELL = "Well"
FTR_ROW = "Row"
FTR_COLUMN = "Column"
FTR_PLATE = "Plate"
ROW_KEYS = ("wellrow", "well_row", "row")
COL_KEYS = ("wellcol", "well_col", "wellcolumn", "well_column", "column", "col")
MEASUREMENTS_GROUP_NAME = "Measurements"
IMAGE_NUMBER = "ImageNumber"
OBJECT_NUMBER = "ObjectNumber"
GROUP_NUMBER = "Group_Number"  # 1-based group index
GROUP_INDEX = "Group_Index"  # 1-based index within group
GROUP_LENGTH = "Group_Length"
R_FIRST_IMAGE_NUMBER = IMAGE_NUMBER + "_" + "First"
R_FIRST_OBJECT_NUMBER = OBJECT_NUMBER + "_" + "First"
R_SECOND_IMAGE_NUMBER = IMAGE_NUMBER + "_" + "Second"
R_SECOND_OBJECT_NUMBER = OBJECT_NUMBER + "_" + "Second"
C_FILE_NAME = "FileName"
C_PATH_NAME = "PathName"
C_URL = "URL"
C_INDEX = "Index"
C_SERIES = "Series"
C_SERIES_NAME = "SeriesName"
C_FRAME = "Frame"
C_FRAMES = "Frames"
C_CHANNEL = "Channel"
C_C = "C"
C_Z = "Z"
C_T = "T"
C_TILE = "TileXYWH"
C_CHANNEL_NAME = "ChannelName"
C_COLOR_FORMAT = "ColorFormat"
C_MONOCHROME = "monochrome"
C_RGB = "RGB"

C_OBJECTS_FILE_NAME = "ObjectsFileName"
C_OBJECTS_PATH_NAME = "ObjectsPathName"
C_OBJECTS_URL = "ObjectsURL"
C_OBJECTS_SERIES = "ObjectsSeries"
C_OBJECTS_SERIES_NAME = "ObjectsSeriesName"
C_OBJECTS_FRAME = "ObjectsFrame"
C_OBJECTS_CHANNEL = "ObjectsChannel"
C_OBJECTS_Z = "ObjectsZPlane"
C_OBJECTS_T = "ObjectsTimepoint"
C_CHANNEL_TYPE = "ChannelType"
C_FILE_LOCATION = "FileLocation"
M_METADATA_TAGS = "_".join((C_METADATA, "Tags"))
M_GROUPING_TAGS = "_".join((C_METADATA, "GroupingTags"))
RESERVED_METADATA_KEYS = (C_URL, C_SERIES, C_SERIES_NAME, C_FRAME, C_FILE_LOCATION,
                          C_COLOR_FORMAT, C_CHANNEL_NAME, C_CHANNEL, C_C, C_Z, C_T, C_TILE)
M_PATH_MAPPINGS = "Path_Mappings"
K_CASE_SENSITIVE = "CaseSensitive"
K_PATH_MAPPINGS = "PathMappings"
K_LOCAL_SEPARATOR = "LocalSeparator"
K_URL2PATHNAME_PACKAGE_NAME = "Url2PathnamePackageName"
F_BATCH_DATA = "Batch_data.mat"
F_BATCH_DATA_H5 = "Batch_data.h5"
C_LOCATION = "Location"
C_NUMBER = "Number"
C_COUNT = "Count"
C_THRESHOLD = "Threshold"
C_PARENT = "Parent"
R_PARENT = "Parent"
C_CHILDREN = "Children"
R_CHILD = "Child"
FTR_CENTER_X = "Center_X"
M_LOCATION_CENTER_X = "%s_%s" % (C_LOCATION, FTR_CENTER_X)
FTR_CENTER_Y = "Center_Y"
M_LOCATION_CENTER_Y = "%s_%s" % (C_LOCATION, FTR_CENTER_Y)
FTR_CENTER_Z = "Center_Z"
M_LOCATION_CENTER_Z = "%s_%s" % (C_LOCATION, FTR_CENTER_Z)
FTR_OBJECT_NUMBER = "Object_Number"
M_NUMBER_OBJECT_NUMBER = "%s_%s" % (C_NUMBER, FTR_OBJECT_NUMBER)
FF_COUNT = "%s_%%s" % C_COUNT
FTR_FINAL_THRESHOLD = "FinalThreshold"
FF_FINAL_THRESHOLD = "%s_%s_%%s" % (C_THRESHOLD, FTR_FINAL_THRESHOLD)
FTR_ORIG_THRESHOLD = "OrigThreshold"
FF_ORIG_THRESHOLD = "%s_%s_%%s" % (C_THRESHOLD, FTR_ORIG_THRESHOLD)
FTR_GUIDE_THRESHOLD = "GuideThreshold"
FF_GUIDE_THRESHOLD = "%s_%s_%%s" % (C_THRESHOLD, FTR_GUIDE_THRESHOLD)
FTR_WEIGHTED_VARIANCE = "WeightedVariance"
FF_WEIGHTED_VARIANCE = "%s_%s_%%s" % (C_THRESHOLD, FTR_WEIGHTED_VARIANCE)
FTR_SUM_OF_ENTROPIES = "SumOfEntropies"
FF_SUM_OF_ENTROPIES = "%s_%s_%%s" % (C_THRESHOLD, FTR_SUM_OF_ENTROPIES)
FF_CHILDREN_COUNT = "%s_%%s_Count" % C_CHILDREN
FF_PARENT = "%s_%%s" % C_PARENT
M_SITE, M_WELL, M_ROW, M_COLUMN, M_PLATE = [
    "_".join((C_METADATA, x))
    for x in (FTR_SITE, FTR_WELL, FTR_ROW, FTR_COLUMN, FTR_PLATE)
]



from ._subscriber import Subscriber


class GridSubscriber(Subscriber):
    """
    A setting that subscribes to grid information providers
    """

    def __init__(
        self,
        text,
        value="Do not use",
        can_be_blank=False,
        blank_text="Leave blank",
        *args,
        **kwargs,
    ):
        super(GridSubscriber, self).__init__(
            text, "gridgroup", value, can_be_blank, blank_text, *args, **kwargs
        )
